import os, sys, datetime, site, pathlib, operator
from collections import OrderedDict, defaultdict, namedtuple
import pandas as pd
import numpy as np
import ete3
import skbio
from scipy import stats
from tqdm import tqdm
from ..io import write_object, read_fasta, read_textfile
from ..utils import pd_dataframe_matmul, format_path, pd_series_collapse, is_dict_like, is_query_class, fragment, assert_acceptable_arguments
from ..transmute.normalization import normalize
from Bio.SeqIO.FastaIO import SimpleFastaParser

__all__ = ["PhylogenomicFunctionalComponents",  "phylogenomically_binned_functional_potential", "mcr_from_directory", "get_taxonomy_lineage_from_identifier", "get_taxonomy_identifier_from_name", "infer_taxonomy", "alpha_diversity", "otu_to_level", "reverse_complement", "count_kmers", "prevalence", "get_intergenic_sequences"]
__all__ = sorted(__all__)

# Extract intergenic sequences
def get_intergenic_sequences(path_annotations:str, path_sequences:str, seq_type="gene", include_peripheral_info=True, peripheral_tag="locus_tag", into=pd.Series):
    """
    Currently tested on GFF3 but it should work for GTF too
    """
    # Get sequences
    sequences = read_fasta(path_sequences, description=False, into=dict, verbose=False)

    # Get gene positions
    d_contig_gene_info = defaultdict(list)

    for i, line in read_textfile(path_annotations):
        if bool(line):
            fields = line.split("\t")
            id_contig = fields[0]
            if fields[2] == "gene":
                pos_start = int(fields[3])-1
                pos_end = int(fields[4])
                data = dict(map(lambda item:item.split("="), filter(bool, fields[-1].split(";"))))
                d_contig_gene_info[id_contig].append((pos_start, pos_end, fields[6], data[peripheral_tag]))

    # Get intergenic sequences
    d_id_seqs= OrderedDict()
    for id_contig, info in d_contig_gene_info.items():
        seq_reference = sequences[id_contig]
        for i in range(0, len(info)-1):
            pos_start = info[i][1]
            pos_end = info[i+1][0]
            id = "{}|intergenic_{}-{}".format(id_contig,pos_start+1, pos_end)
            if include_peripheral_info:
                id = "{}|({}){}|({}){}".format(id,info[i][2], info[i][3], info[i+1][2], info[i+1][3])
            seq_intergenic = seq_reference[pos_start:pos_end]
            if len(seq_intergenic) > 0:
                d_id_seqs[id] = seq_intergenic
    return into(d_id_seqs)


# Phylogeneomically binned functional potential
def phylogenomically_binned_functional_potential(X:pd.DataFrame,MCR:pd.DataFrame):
    """
    X: DataFrame of scaled compositional data that preserves proportionality (e.g. TPM or relative-abundance)
    MCR: DataFrame

    X.shape (n,m)
    MCR.shape (m,p)
    Output: (n,p)
    """
    assert X.shape[1] == MCR.shape[0], "Shapes are not compatible"
    return pd_dataframe_matmul(X,MCR)

# Parse the files from KEGG Maple
def mcr_from_directory(directory:str, name=None, end_tag="module.xls", mcr_field="MCR % (ITR)", index_name_bins = "id_bin", index_name_module = "id_module", store_raw_data=False):
    """
    directory:
    Path to kegg maple results which should have the following structure:
    | directory
    | | bin_1
    | | | xyz1.module.xls
    | | bin_2
    | | | xyz2.module.xls

    Returns:
    namedtuple with the following attributes:
        .ratios = MCR ratio
        .metadata = module metadata
        .name = user-provided name of experiment
        .raw_data = {id_bin:raw_kegg_maple_}

    Assumes subdirectories are bin_ids and
    each directory has a file that has the
    following format: 8206709269.module.xls

    https://www.genome.jp/maple-bin/mapleSubmit.cgi?aType=sDirect
    """
    MCR = namedtuple("MCR", ["ratios", "metadata", "name", "raw_data"])
    # Format the path
    directory = format_path(directory)
    # Convert the path to a Path object
    directory = pathlib.Path(directory)

    # Check directory
    assert directory.exists(), f"{str(directory.absolute())} does not exist"
    assert directory.is_dir(), f"{str(directory.absolute())} is not a directory"

    # Iterate through each KEGG Maple result dataset
    d_bin_mcr = OrderedDict()
    module_metadata = list()
    d_bin_raw = dict()
    for subdir in tqdm(directory.iterdir(), f"Reading subdirectories in {str(directory)}"):
        sheets = list()
        if subdir.is_dir():
            id_bin = subdir.name
            for file in subdir.iterdir():
                if file.name.endswith(end_tag):
                    for sheetname in ["module_pathway", "module_complex", "module_function", "module_signature"]:
                        df = pd.read_excel(file, index_col=0, sheetname=sheetname).set_index("ID", drop=False)
                        sheets.append(df)
            df_bin = pd.concat(sheets, axis=0)
            module_metadata.append(df_bin.loc[:,["Type", "Small category", "Name(abbreviation)"]])
            d_bin_mcr[id_bin] = df_bin[mcr_field]
            # Store raw data
            if store_raw_data:
                d_bin_raw[id_bin] = df_bin.drop("ID", axis=1)
    # Module Completion Ratios
    df_mcr = pd.DataFrame(d_bin_mcr).T.fillna(0.0)
    df_mcr.index.name = index_name_bins
    df_mcr.columns.name = index_name_module

    # Metadata
    df_metadata = pd.concat(module_metadata, axis=0).drop_duplicates() # May not be the most efficient but w/e

    return MCR(ratios=df_mcr, metadata=df_metadata, name=name, raw_data=d_bin_raw)


# Phylogenomic Functional Components
class PhylogenomicFunctionalComponents(object):
    __doc__ = """
    PhylogenomicFunctionalComponents
    ================================
    This includes methods to create Phylogenomic Functional Components.
    For `components_taxonomy` check out the databases available with Soothsayer:
    %s/soothsayer/db/

    Formatting should resemble the following:

    components_taxonomy.head()
    NODE_100000_length_1286_cov_38.26_32_868_-            UNBINNED__Streptococcus mutans
    NODE_100000_length_1286_cov_38.26_996_1285_-           UNBINNED__Streptococcus mitis
    NODE_100001_length_1286_cov_31.4249_1_152_+               UNBINNED__Neisseria mucosa
    NODE_100001_length_1286_cov_31.4249_257_925_+             UNBINNED__Neisseria mucosa
    NODE_100004_length_1286_cov_2.62551_1_686_-      UNBINNED__Leptotrichia goodfellowii
    Name: taxonomy, dtype: object

    components_functional.head() # Note this one will most likely have multiple functions for each ORF which is why it's a set and not a string
    NODE_100009_length_1286_cov_12.6458_1_870_+         {M00432__Leucine biosynthesis, 2-oxoisovalerate => 2-oxoisocaproate}
    NODE_395196_length_426_cov_39.5364_1_426_+       {M00051__Uridine monophosphate biosynthesis, glutamine (+ PRPP) => UMP}
    NODE_395187_length_426_cov_2.6442_1_426_-                         {M00120__Coenzyme A biosynthesis, pantothenate => CoA}
    NODE_39517_length_2534_cov_7.92457_605_1108_+                                             {M00063__CMP-KDO biosynthesis}
    NODE_39517_length_2534_cov_7.92457_1_572_+        {M00018__Threonine biosynthesis, aspartate => homoserine => threonine}
    """%site.getsitepackages()[0]
    def __init__(self,
                 # Input data
                 lengths:pd.Series,
                 components_taxonomy:pd.Series,
                 components_functional:pd.Series,
                 # Labels
                 label_unknown:str="unknown_function_collective",
                 attr_type="ORF",
                 obsv_type=None,
                 functional_type=None,
                 name=None,
                 description=None,
                 # Misc
                 delimiter = "|",
                ):
        # Input data
        self.lengths = lengths
        self.components_taxonomy = components_taxonomy
        self.components_functional = components_functional
        # Description
        self.name = name
        self.description = description
        # Data Types
        self.attr_type = attr_type
        self.obsv_type = obsv_type
        self.functional_type = functional_type
        # Miscellaneous
        self.label_unknown = label_unknown
        self.delimiter = delimiter
        self.__synthesized__ = datetime.datetime.utcnow()
        # ==========
        # Initialize
        # ==========
        print(f"Phylogenomic Functional Components",
               "==================================",
               sep="\n",
               file=sys.stderr,
         )
        # Assign ORFs to PGFCs
        self.pgfc_collection = self._assign_orfs_to_pgfcs()
        # Compile metadata for each PGFC
        self.metadata = self._get_metadata()
        # Datasets
        self.datasets = OrderedDict()
    def __repr__(self):
        class_name = str(self.__name__)
        return f"{class_name}(name:{self.name}, num_datasets:{len(self.datasets)}, num_pgfcs:{self.metadata.shape[0]})"

    # Assign ORFS to each PGFC
    def _assign_orfs_to_pgfcs(self):
        pgfc_collection = defaultdict(list)
        for taxonomy_component, Se_data in tqdm(self.components_taxonomy.groupby(self.components_taxonomy), f"Assigning ORFs"):
            idx_orfs_for_taxonomycomponent = set(Se_data.index)
            # Process missing functions
            idx_orfs_with_function = set(self.components_functional.index)
            idx_orfs_no_function = idx_orfs_for_taxonomycomponent - idx_orfs_with_function
            if len(idx_orfs_no_function) > 0:
                id_pgfc = self.delimiter.join([taxonomy_component, self.label_unknown])
                pgfc_collection[id_pgfc] = list(idx_orfs_no_function)
                del id_pgfc # Reset the namespace
            # Process those with functional category
            for id_orf, functional_list in self.components_functional[idx_orfs_for_taxonomycomponent & idx_orfs_with_function].iteritems():
                for functional_component in functional_list:
                    id_pgfc = self.delimiter.join([taxonomy_component, functional_component])
                    pgfc_collection[id_pgfc].append(id_orf)
        # Convert list to non-redundant sets
        return {id_pgfc:set(orfs) for id_pgfc, orfs in pgfc_collection.items()}

    # Get metadata for PGFCs
    def _get_metadata(self):
        metadata = defaultdict(dict)
        for id_pgfc, orfs in tqdm(self.pgfc_collection.items(), "Compiling metadata"):
            # Length
            query_lengths = self.lengths[orfs].values
            try:
                k2, p = stats.normaltest(query_lengths)
            except ValueError:
                k2 = p = np.nan
            metadata[id_pgfc]["orfs"] = orfs
            metadata[id_pgfc]["number_of_orfs"] = len(query_lengths)
            metadata[id_pgfc]["lengths"] = list(query_lengths)
            metadata[id_pgfc]["µ(lengths)"] = query_lengths.mean()
            metadata[id_pgfc]["∑(lengths)"] = query_lengths.sum()
            metadata[id_pgfc]["sem(lengths)"] = stats.sem(query_lengths)
            metadata[id_pgfc]["var(lengths)"] = np.var(query_lengths)
            metadata[id_pgfc]["normaltest__statistic"] = k2
            metadata[id_pgfc]["normaltest__pvalue"] = p
            metadata[id_pgfc]["taxonomy_component"] = id_pgfc.split(self.delimiter)[0]
            if id_pgfc.endswith(self.label_unknown):
                metadata[id_pgfc]["functional_component"] = np.nan
            else:
                metadata[id_pgfc]["functional_component"] = id_pgfc.split(self.delimiter)[-1]
        # Metadata
        df_meta = pd.DataFrame(metadata).T.loc[:,["taxonomy_component", "functional_component","number_of_orfs", "∑(lengths)", "µ(lengths)","sem(lengths)","var(lengths)","normaltest__statistic", "normaltest__pvalue", "lengths", "orfs"]].sort_values(["taxonomy_component", "functional_component", "number_of_orfs"], ascending=[True, True, False])
        df_meta.index.name = "id_phylogenomic-functional-component"

        # Ints
        for id_field in ["number_of_orfs", "∑(lengths)"]:
            df_meta[id_field] = df_meta[id_field].astype(int)
        # Floats
        for id_field in ["µ(lengths)","sem(lengths)","var(lengths)","normaltest__statistic", "normaltest__pvalue"]:
            df_meta[id_field] = df_meta[id_field].astype(float)
        return df_meta

    # Compute the summed ORF counts for each PGFC
    def compute_dataset(self, df_orfs:pd.DataFrame, name=None, copy_data=True, return_dataset=True, **attr):
        assert all(set(self.components_taxonomy.index) <= query_gene_set for query_gene_set in [set(self.lengths.index), set(df_orfs.columns)]), f"All genes in `components_taxonomy` must be in `lengths` and `df_orfs`"
        assert set(df_orfs.columns) <= set(self.lengths.index), f"All genes in `df_orfs` must be in `lengths`"
        components_taxonomy = self.components_taxonomy.dropna()
        components_functional = self.components_functional.dropna()

        # Name
        if name is None:
            name = len(self.datasets)

        # Sum ORFs for each PGFC
        d_pgfc_counts = dict()
        for id_pgfc, orfs in tqdm(self.pgfc_collection.items(), f"Summing ORFs for dataset = `{name}`"):
            d_pgfc_counts[id_pgfc] = df_orfs.loc[:,orfs].sum(axis=1) # Removed set construction here because orfs are non-redundant from _assign_orfs_to_pgfcs
        df_pgfc = pd.DataFrame(d_pgfc_counts).sort_index(axis=1)

        # Store dataset
        if copy_data:

            self.datasets[name] = {
                "shape_input":df_orfs.shape,
                "counts":df_pgfc,
                "shape_output":df_pgfc.shape,
                **attr,
            }
        if return_dataset:
            return df_pgfc
        else:
            return self
    # Get PGFC counts
    def get_dataset(self, name):
        assert name in self.datasets, f"{name} not in `datasets`.  Try`compute_dataset` and repeat."
        return self.datasets[name]["counts"]
    # Write object to file
    def to_file(self, path:str, compression="infer"):
        write_object(self, path=path, compression=compression)
        return self

# Get taxonomy lineage
def get_taxonomy_lineage_from_identifier(identifiers, name=None, translate_ids=True, ncbi:ete3.NCBITaxa=None, verbose=0, mode="infer", include_taxid=True, include_levels=["id_taxon", "phylum", "class", "order", "family", "genus", "species"]):
    """
    Input: Single Taxonomy ID or a collection of identifiers w/ {id_orf:id_taxon}
    Output: pd.Series/pd.DataFrame of rank and taxonomy labels
    modes: {'infer', 'batch', 'singular'}
    """
    accepted_modes = {'batch', 'singular'}
    verbose = int(verbose)
    # Get database
    if ncbi is None:
        ncbi = ete3.NCBITaxa()

    # Infer mode
    if mode == "infer":
        if is_dict_like(identifiers):
            mode = "batch"
        else:
            mode = "singular"
        if verbose > 0:
            print(f"Inferred mode: {mode}", file=sys.stderr)
        assert mode != "infer", "Cannot infer `mode`.  Please explicitly provide mode."

    # Singular
    if mode == "singular":
        # Handle missing data
        if pd.isnull(identifiers):
            return pd.Series([], name=name)

        # Access the database
        try:
            id_taxon = int(identifiers)
            if name is None:
                name = id_taxon
            lineage = ncbi.get_lineage(id_taxon)
            ranks = dict(filter(lambda x: x[1] != "no rank", ncbi.get_rank(lineage).items()))
            Se_taxonomy = pd.Series(ncbi.get_taxid_translator(ranks.keys()), name=name)
            if translate_ids:
                Se_taxonomy.index = Se_taxonomy.index.map(lambda x:ranks[x])
            return Se_taxonomy

        # Handle taxonomy IDs that are not in database
        except ValueError:
            if verbose > 1:
                print(id_taxon, file=sys.stderr)
            return pd.Series([], name=name)
    # Batch
    if mode == "batch":
        if not is_query_class(identifiers, "Series"):
            identifiers = pd.Series(identifiers)

        # Group each taxonomy identifier
        dataframes = list()
        for id_taxon, group in tqdm(pd_series_collapse(identifiers).iteritems(), "Searching lineage from taxonomy identifier"):
            number_of_orfs_in_group = len(group)
            Se_taxonomy = get_taxonomy_lineage_from_identifier(identifiers=id_taxon, name=None, translate_ids=translate_ids, ncbi=ncbi, verbose=verbose, mode="singular")
            df_taxon = pd.DataFrame(number_of_orfs_in_group*[Se_taxonomy])
            df_taxon.index = group
            if include_taxid:
                df_taxon["id_taxon"] = id_taxon
            dataframes.append(df_taxon)
        df_collection = pd.concat(dataframes, axis=0)
        if include_levels is None:
            include_levels = df_collection.columns
        else:
            include_levels = [*filter(lambda level:level in df_collection.columns, include_levels)]

        idx_missing_orfs = set(identifiers.index) - set(df_collection.index)
        number_of_orfs_missing = len(idx_missing_orfs)
        if number_of_orfs_missing > 0:
            A = np.empty((number_of_orfs_missing, df_collection.shape[1]))
            A[:] = np.nan
            df_missing = pd.DataFrame(A, index=idx_missing_orfs, columns=df_collection.columns)
            return pd.concat([df_collection, df_missing], axis=0).loc[identifiers.index, include_levels]
        else:
            return df_collection.loc[identifiers.index, include_levels]

# Get taxonomy identifiers from string name_taxonomy
def get_taxonomy_identifier_from_name(name_taxonomy, mode="infer", ncbi=None, into=pd.Series):
    """
    name_taxonomy can be either a single string or iterable of strings
    """
    accepted_modes = {'batch', 'singular'}

    # Get database
    if ncbi is None:
        ncbi = ete3.NCBITaxa()

    if mode == "infer":
        if not is_nonstring_iterable(name_taxonomy):
            mode = "singular"
        else:
            mode = "batch"
        assert mode != "infer", "Cannot infer `mode`.  Please explicitly provide mode."

    if mode == "singular":
        name_taxonomy = [name_taxonomy]
    name_taxonomy = [*map(lambda x:x.strip(),name_taxonomy)]
    Se_taxonomy = pd.Series(ncbi.get_name_translator(name_taxonomy)).sort_index().map(lambda x:x[0])
    if mode == "singular":
        return Se_taxonomy.values[0]
    if mode == "batch":
        idx_missing = set(name_taxonomy) - set(Se_taxonomy.index)
        number_of_name_taxonomy_missing = len(idx_missing)
        if number_of_name_taxonomy_missing > 0:
            Se_missing = pd.Series([np.nan]*number_of_name_taxonomy_missing, index=idx_missing)
            Se_taxonomy = pd.concat([Se_taxonomy, Se_missing])
        return into(Se_taxonomy)

# Infer taxonomy for a sequence grouping
def infer_taxonomy(
        grouping_sequence:pd.Series,
        grouping_taxonomy:pd.Series,
        percent_identities:pd.Series,
        grouping_protein_domains:pd.Series=None,
        include_protein_domains=None
    ):
    """
    Format:
    group_sequence:pd.Series where {id_orf:contig-or-bin-identifier}
    grouping_taxonomy:pd.Series where {id_orf:taxonomy-label}
    percent_identities:pd.Series where {id_orf:percent-identity}
    grouping_protein_domains:pd.Series where {id_orf:[PFAM_1, TIGRFAM_2, etc.]} (Optional)
    include_protein_domains: iterable of target protein-domain names (Optional)
    """
    # Comparable set
    idx_comparable = grouping_sequence.index & grouping_taxonomy.index & percent_identities.index

    # Subset data based on HMMs
    if include_protein_domains is not None:
        assert grouping_protein_domains is not None, "If `include_protein_domains` is provided then you must also provide `grouping_protein_domains`"
        include_protein_domains = set(include_protein_domains)
        grouping_protein_domains = grouping_protein_domains[grouping_protein_domains.map(bool)]
        mask_include_protein_domains = grouping_protein_domains.map(lambda x: bool(set(x) & include_protein_domains))
        grouping_protein_domains = grouping_protein_domains[mask_include_protein_domains]
        idx_comparable = idx_comparable & grouping_protein_domains.index

    # Convert percent identities to ratios
    if percent_identities.max() > 1.0:
        percent_identities = percent_identities/100

    # Infer taxonomy
    d_seqgroup_hits = defaultdict(lambda: defaultdict(int))
    d_seqgroup_weights = defaultdict(lambda: defaultdict(float))

    data = pd.DataFrame([grouping_sequence, grouping_taxonomy, percent_identities]).T.loc[idx_comparable,:].values

    for i, (id_orf, row_data) in tqdm(enumerate(zip(idx_comparable, data)), "Summing weights from ORFs"):
        id_seqgroup, id_taxon, p_identity = row_data
        d_seqgroup_hits[id_seqgroup][id_taxon] += 1
        d_seqgroup_weights[id_seqgroup][id_taxon] += p_identity

    # Calculate scores
    d_seqgroup_inferred = defaultdict(dict)
    for id_seqgroup, d_taxa_weight in tqdm(d_seqgroup_weights.items(), "Scoring taxonomy for each sequence group"):
        w = pd.Series(d_taxa_weight, name=id_seqgroup)
        scores = w/w.sum()
        d_seqgroup_inferred[id_seqgroup]["weights"] = sorted(w.iteritems(), key=lambda x:x[-1], reverse=True)
        max_score = scores.max()
        d_seqgroup_inferred[id_seqgroup]["score"] = max_score
        if max_score < 1:
            max_scoring_taxa = scores.compress(lambda x: x == max_score).index
            if max_scoring_taxa.size > 1:
                inferred_taxonomy = list(max_scoring_taxa) # Added 2018-11-30
            else:
                inferred_taxonomy = max_scoring_taxa[0]
        else:
            inferred_taxonomy = scores.index[0]
        d_seqgroup_inferred[id_seqgroup]["inferred_taxonomy"] = inferred_taxonomy
        d_seqgroup_inferred[id_seqgroup]["number_of_total_hits"] = sum(d_seqgroup_hits[id_seqgroup].values())

    df_inferred = pd.DataFrame(d_seqgroup_inferred).T.loc[:,[ "inferred_taxonomy", "score", "number_of_total_hits", "weights"]]
    df_inferred.index.name = "id_seqgroup"
    return df_inferred

# Group Otus by higher taxonomy level
def otu_to_level(X:pd.DataFrame, taxonomy:pd.DataFrame, level):
    """
    X => pd.DataFrame of Otu counts with columns as Otus and rows as samples
    taxonomy => pd.DataFrame of taxonomy assignments for Otus
    level => a taxonomy level in the `taxonomy` dataframe
    """
    return X.groupby(taxonomy[level], axis=1).sum()

# Alpha diversity calculations
def alpha_diversity(X:pd.DataFrame, metric="richness", taxonomy:pd.DataFrame=None, mode="infer", idx_taxonomy=['Phylum', 'Class', 'Order', 'Family', 'Genus', 'Species'],  name=None, base=2, obsv_type=None, **alpha_kws):
    """
    X => pd.DataFrame of Otu counts with columns as Otus and rows as samples
    metric => a callable or a string {entropy, richness, gini, singletons}
    taxonomy => pd.DataFrame of taxonomy assignments for Otus
    """
    # Alpha measures
    def _entropy(x, **alpha_kws):
        return stats.entropy(x, base=base, **alpha_kws)
    def _richness(x, **alpha_kws):
        return (x > 0).sum().astype(int)
    def _gini(x, **alpha_kws):
        return skbio.diversity.alpha.gini_index(x, **alpha_kws)
    def _singletons(x, **alpha_kws):
        return (x == 1).sum().astype(int)
    d_metric_fn = {"entropy":_entropy, "richness":_richness, "gini":_gini, "singletons":_singletons}

    # Supported diversity measures
    if hasattr(metric, "__call__"):
        func = metric
    else:
        supported_metrics =  list(d_metric_fn.keys())
        assert metric in supported_metrics, f"`{metric}` is not compatible.  Only available alpha diversity measures are {supported_metrics}"
        func =  d_metric_fn[metric]
        name = metric

    # Compute diversity
    if mode == "infer":
        mode = "batch" if taxonomy is not None else "singular"

    assert mode in {"singular", "batch"}, "Please specify either 'singular', 'batch', or 'infer' for the mode"
    if mode == "singular":
        Se_alpha = X.apply(lambda x: func(x, **alpha_kws), axis=1)
        Se_alpha.index.name = name
        return Se_alpha

    if mode == "batch":
        assert taxonomy is not None, "`taxonomy` cannot be `None` when `mode='batch'`"
        d_level_metric = OrderedDict()
        for level in idx_taxonomy:
            if level in taxonomy.columns:
                df_level = otu_to_level(X, taxonomy, level=level)
                d_level_metric[level] = alpha_diversity(df_level, metric=metric, taxonomy=None, mode="singular", base=base, **alpha_kws)
            else:
                print(f"Skipping taxonomy level `{level}` because it is not in the taxonomy dataframe", file=sys.stderr)

        d_level_metric["Otu"] = alpha_diversity(X, metric=metric, taxonomy=None, mode="singular", base=base, **alpha_kws)
        df_level_metric = pd.DataFrame(d_level_metric)
        df_level_metric.index.name = f"id_{obsv_type}"
        return df_level_metric

# Reverse Complement
def reverse_complement(seq:str):
    """
    Inputs a sequence of DNA and returns the reverse complement se$
    Outputs a reversed string of the input where A>T, T>A, G>C, an$
    """
    conversion = str.maketrans('ACGTacgt','TGCAtgca')
    comp = seq.translate(conversion)
    rev_comp = comp[::-1]
    return rev_comp

# K-mer frequency counter
def count_kmers(sequences, K=5, frequency=True, collapse_reverse_complement=False, pseudocount=1, fill_missing_kmers=False):
    """
    sequences: (1) a sequence string, (2) a dict_like of (id,seq) pairs, or (3) path to fasta
    collapse_reverse_complement reduces data dimensionality
    """
    def _infer_mode(sequences):
        mode = None
        if isinstance(sequences, str):
            if not os.path.exists(sequences):
                return "string_sequence"
            if os.path.exists(sequences):
                return "filepath"
        if is_dict_like(sequences):
            return "dict_like"
        assert mode is not None, "Unrecognized format:  Please provide: (1) a sequence string, (2) a dict_like of (id,seq) pairs, or (3) path to fasta"


    # Get mapping to collapse a strand
    def _collapse(kmers):
        kmer_collapse = dict()
        for kmer in kmers:
            r_kmer = reverse_complement(kmer)
            if r_kmer in kmer_collapse:
                kmer_collapse[kmer] = r_kmer
            else:
                kmer_collapse[kmer] = kmer
        return kmer_collapse
    def _count(seq:str, K=5, frequency=True, collapse_reverse_complement=False,  pseudocount=1, name=None):
        """
        Collapse only works for nondegenerate kmers
        """
        seq = seq.upper()
        alphabet = {"A", "C", "G", "T"}
        query_alphabet = set(list(seq))
        # degenerate_kmers = {'B', 'D', 'H', 'K', 'M', 'N', 'R', 'S', 'V', 'W', 'Y'}
        assert query_alphabet <= alphabet, "This cannot be used with degenerate characters"

        # Count kmers
        kmer_counts = defaultdict(int)
        for kmer in fragment(seq=seq, K=K, step=1, overlap=True):
            kmer_counts[kmer] += 1
        kmer_counts = pd.Series(kmer_counts, name=name) + pseudocount

        # Collapse double stranded
        if collapse_reverse_complement:
            kmer_collapse = _collapse(sorted(kmer_counts.index))
            kmer_counts = kmer_counts.groupby(kmer_collapse).sum()

        # Normalize
        kmer_counts = kmer_counts + pseudocount
        if frequency:
            return kmer_counts/kmer_counts.sum()
        if not frequency:
            return kmer_counts

    # Infer mode
    mode = _infer_mode(sequences)

    # Single sequence (string)
    if mode == "string_sequence":
        return _count(seq, K=K, frequency=frequency, collapse_reverse_complement=collapse_reverse_complement, pseudocount=pseudocount)
    else:
        # Multiple sequences
        params = dict(K=K, frequency=False, collapse_reverse_complement=False, pseudocount=0)
        f_sequences = None
        # Dict-like
        if mode == "dict_like":
            iterable = sequences.items()

        # Read in files
        if mode == "filepath":
            f_sequences = open(sequences, "r")
            iterable = SimpleFastaParser(f_sequences)

        # Count kmers
        data = list()
        for id, seq in tqdm(iterable, "Counting k-mers"):
            data.append(_count(name=id, seq=seq, **params))
        # Close file handle if necessary
        if f_sequences is not None:
            f_sequences.close()
        df_kmer_counts = pd.DataFrame(data)

        # Fill empty
        if fill_missing_kmers:
            missing_kmers = set(["".join(kmer) for kmer in itertools.product("ATCG",repeat=K)]) - set(df_kmer_counts.columns)
            A = np.zeros((df_kmer_counts.shape[0], len(missing_kmers)))
            df_missing = pd.DataFrame(A, index=df_kmer_counts.index, columns=list(missing_kmers))
            df_kmer_counts = pd.concat([df_kmer_counts, df_missing], axis=1)

        # Sort K-mers and fill null
        df_kmer_counts = df_kmer_counts.sort_index(axis=1, ascending=True).fillna(0)

        # Collapse Double-Stranded
        if collapse_reverse_complement:
            kmer_collapse = _collapse(df_kmer_counts.columns)
            df_kmer_counts = df_kmer_counts.groupby(kmer_collapse, axis=1).sum()

        # Pseudocount
        df_kmer_counts = df_kmer_counts + pseudocount

        # Normalize
        if frequency:
            return normalize(df_kmer_counts, method="tss", axis=1)
        else:
            return df_kmer_counts.astype(int)

# Prevalence
def prevalence(X:pd.DataFrame, interval_type="closed", name=None, scale=False) -> pd.Series:
    # Dimensions
    n,m = X.shape

    assert_acceptable_arguments(query=[interval_type],target=["closed", "open"], operation="le")

    # Attribtue depth
    data = (X > 0).sum(axis=0)

    # Comparison operation
    operation = {"closed":operator.ge, "open":operator.gt}[interval_type]

    # Number of attributes for tolerance
    d_tol_numattrs = OrderedDict()
    for tol in range(1,n+1):
        d_tol_numattrs[tol] = data.map(lambda x: operation(x,tol)).sum()

    # Prevalence
    x_prevalence = pd.Series(d_tol_numattrs, name=name)

    # Scale the prevalence
    if scale:
        x_prevalence.index = x_prevalence.index/n
    return x_prevalence



from .r_wrappers import *
from .packages import dynamicTreeCut
from .packages import WGCNA
from .packages import philr
from .packages import edgeR
from .packages import metagenomeSeq

__all__ = ["dynamicTreeCut", "WGCNA", "philr", "edgeR", "metagenomeSeq" ,"R_package_retrieve", "R_packages"]
__all__ = sorted(__all__)



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time
from collections import OrderedDict, defaultdict
from io import StringIO

# PyData
import pandas as pd
import numpy as np
import networkx as nx

# Tree
import ete3
import skbio

# SciPy
from scipy import stats
from scipy.cluster import hierarchy as sp_hierarchy
from scipy.spatial.distance import squareform
try:
    from fastcluster import linkage
except ImportError:
    from scipy.cluster.hierarchy import linkage
    print("Could not import `linkage` from `fastcluster` and using `scipy.cluster.hierarchy.linkage` instead", file=sys.stderr)

# Soothsayer
from ..utils import dict_reverse

__all__ = ["dism_to_linkage", "linkage_to_newick", "dism_to_tree", "name_ete_nodes", "ete_to_skbio", "nx_to_ete", "ete_to_nx"]

# Create clusters
def dism_to_linkage(df_dism, method="ward"):
    """
    Input: A (m x m) dissimilarity Pandas DataFrame object where the diagonal is 0
    Output: Hierarchical clustering encoded as a linkage matrix

    Further reading:
    http://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.cluster.sp_hierarchy.linkage.html
    https://pypi.python.org/pypi/fastcluster
    """
    #Linkage Matrix
    dist = squareform(df_dism.values)
    return linkage(dist,method=method)

# Convert dendrogram to Newick
def linkage_to_newick(Z, labels):
    """
    Input :  Z = linkage matrix, labels = leaf labels
    Output:  Newick formatted tree string
    Edit: 2018-June-28
    https://github.com/biocore/scikit-bio/issues/1579
    """
    tree = sp_hierarchy.to_tree(Z, False) #scipy.sp_hierarchy.to_tree
    def build_newick(node, newick, parentdist, leaf_names):
        if node.is_leaf():
            return f"{leaf_names[node.id]}:{(parentdist - node.dist)/2}{newick}"
        else:
            if len(newick) > 0:
                newick = f"):{(parentdist - node.dist)/2}{newick}"
            else:
                newick = ");"
            newick = build_newick(node.get_left(), newick, node.dist, leaf_names)
            newick = build_newick(node.get_right(), f",{newick}", node.dist, leaf_names)
            newick = f"({newick}"
            return newick
    return build_newick(tree, "", tree.dist, labels)

# Dissimilarity to ete
def dism_to_tree(df_dism:pd.DataFrame, into=ete3.Tree, method="ward", name=None):
    """
    https://en.wikipedia.org/wiki/Newick_format
    Input: pd.DataFrame distance/dissimilarity matrix, diagonal must be 0, must be symmetrical
    Output: ete3 Tree
    """
    # Check labels for tuples
    labels = df_dism.index
    encrypt=False
    if any(map(lambda x:type(x) == tuple, labels)):
        d_encode = dict(zip(labels, map(str,range(len(labels)))))
        d_decode = dict_reverse(d_encode)
        df_dism.index = df_dism.columns = labels = df_dism.index.map(lambda x: d_encode[x])
        encrypt = True
    # Linkage
    Z = linkage(df_dism, method=method)
    # Newick
    newick = linkage_to_newick(Z, labels=labels)
    tree = into(newick=newick)
    tree.name= name
    # Decode
    if encrypt:
        for node in tree.traverse():
            if node.is_leaf():
                node.name = d_decode[node.name]
    return tree

# Name ete3 nodes
def name_ete_nodes(tree,  node_prefix="y"):
    intermediate_node_index = 1
    for node in tree.traverse():
        if not node.is_leaf():
            node.name = f"{node_prefix}{intermediate_node_index}"
            intermediate_node_index += 1
    return tree

# Convert ete3 to skbio
def ete_to_skbio(tree, node_prefix="y"):
    if node_prefix is not None:
        tree = name_ete_nodes(tree, node_prefix=node_prefix)
    return skbio.TreeNode.read(StringIO(tree.write(format=1, format_root_node=True)))

# Convert a NetworkX graph object to an ete3 Tree
def nx_to_ete(G, deepcopy=True, into=ete3.Tree, root="infer", verbose=False):
    # Add nodes
    subtrees = OrderedDict()
    for node_data in G.nodes(data=deepcopy):
        if deepcopy:
            node, node_attrs = node_data
            subtree = into()
            subtree.name = node # Had to do this becuase ClusterNode doesn't have `name` attribute in args
            subtree.add_features(**node_attrs)
        else:
            node = node_data
            subtree = into()
            subtree.name = node # Had to do this becuase ClusterNode doesn't have `name` attribute in args
        subtrees[node] = subtree

    # Add edges
    for edge_data in G.edges(data=deepcopy):
        if deepcopy:
            parent, child, attrs = edge_data
            subtrees[parent].add_child(child=subtrees[child])
            subtrees[parent].add_features(**{(child,k):v for k,v in node_attrs})
            subtrees[child].add_features(**{(parent,k):v for k,v in node_attrs})
        else:
            parent, child, attrs = edge_data
            subtrees[parent].add_child(child=subtrees[child])
    # Inferring the tree root
    if root == "infer":
        descedants = sorted(map(lambda subtree: (subtree, len(subtree.get_descendants())), subtrees.values()), key=lambda x:x[1])
        descendant_counts = list(map(lambda x:x[1], descedants))
        tree_root, num_descendants = descedants[-1]
        if descendant_counts.count(num_descendants) > 1:
            root = None
            if verbose:
                print(f"Warning: The root node could not be inferred because no node has a maximum number of descendants", file=sys.stderr )
        else:
            if verbose:
                print(f"{tree_root.name} was inferred as root with n={num_descendants} descendants", file=sys.stderr )
            return tree_root
    if root is not None:
        if root not in subtrees:
            if verbose:
                print(f"{tree_root.name} was not in the networkx graph {G.name}", file=sys.stderr )
            root = None
        else:
            return subtrees[root]
    # Returning dictionary of subtrees
    if root is None:
        return subtrees

# ete3 to NetworkX
def ete_to_nx(tree, create_using=None, node_prefix="infer"):
    # Get graph
    if create_using is None:
        graph = nx.OrderedDiGraph()
    else:
        graph = create_using
    # Add node prefixes
    if node_prefix == "infer":
        if "" in tree:
            node_prefix = "y"
        else:
            node_prefix = None
    if node_prefix is not None:
        tree = name_ete_nodes(tree, node_prefix=node_prefix)
    else:
        assert "" not in tree, "All nodes (including intermediate) must be named"
    # Construct graph
    for node in tree.traverse():
        for child in node.get_children()[::-1]:
            branch_length = tree.get_distance(node, child)
            graph.add_edge(node.name, child.name, weight=branch_length)
    return graph



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys, time

import pandas as pd

# Soothsayer
from ..r_wrappers import *
from soothsayer.symmetry import *
from soothsayer.utils import *

# ==============================================================================
# R Imports
# ==============================================================================
from rpy2 import robjects, rinterface
from rpy2.robjects.packages import importr
from rpy2.rinterface import RRuntimeError
from rpy2.robjects import pandas2ri
pandas2ri.activate()
R = robjects.r
NULL = robjects.rinterface.NULL
rinterface.set_writeconsole_regular(None)

# R packages
metagenomeSeq = R_package_retrieve("metagenomeSeq")

# CSS Normalization
def normalize_css(X):
    """
    metagenomeSeq: https://github.com/HCBravoLab/metagenomeSeq
    Tranposes pd.DataFrame object | attributes = columns --> attributes = rows
    Calculates NormFactors via metagenomeSeq
    Scales counts with scaling factors
    """
    assert isinstance(X, pd.DataFrame), "type(X) must be pd.DataFrame"
    # metagenomeSeq = R_package_retrieve("metagenomeSeq")
    # Run R-Pipeline
    def run_css(X):
        # Convert pd.DataFrame to R-object
        rX = pandas2ri.py2ri(X)
        # Instantiate metagenomeSeq experiment
        obj = metagenomeSeq.newMRexperiment(rX)
        # Calculate NormFactors
        p = metagenomeSeq.cumNormStatFast(obj)
        obj = metagenomeSeq.cumNorm(obj, p=p)#     # Calculate Library Size
        # CSS Normalization
        rDF_css = metagenomeSeq.MRcounts(obj, norm = True, log = False)
        return pandas2ri.ri2py(rDF_css)

    # Gene axis as rows
    X = X.T

    # Labels
    idx_attrs = X.index
    idx_obsvs = X.columns

    # Compute CSS
    df_css = pd.DataFrame(run_css(X), index=idx_attrs, columns=idx_obsvs)

    # Adjust attribute axis
    return df_css.T



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, copy, pathlib, importlib
from collections import *

# Compression & Serialization
import pickle, gzip, bz2, zipfile

# Soothsayer
from soothsayer.utils import format_path, infer_compression

# ======
# Future
# ======
# (1) Add ability for `path` arguments to be `pathlib.Path`
# (2) Add ability to process ~ in path

# =============
# Serialization
# =============
# Writing serial object
def write_object(obj, path:str, compression="infer", serialization_module=pickle, protocol=pickle.HIGHEST_PROTOCOL, *args):
    """
    Extensions:
    pickle ==> .pkl
    dill ==> .dill
    gzipped-pickle ==> .pgz
    bzip2-pickle ==> .pbz2
    """
    assert obj is not None, "Warning: `obj` is NoneType"
    path = format_path(path, str)

    # Use infer_compression here
    if compression == "infer":
        _ , ext = os.path.splitext(path)
        if (ext == ".pkl") or (ext == ".dill"):
            compression = None
        if ext in {".pgz", ".gz"}:
            compression = "gzip"
        if ext in {".pbz2", ".bz2"}:
            compression = "bz2"
    if compression is not None:
        if compression == "bz2":
            f = bz2.BZ2File(path, "wb")
        if compression == "gzip":
            f = gzip.GzipFile(path, "wb")
    else:
        f = open(path, "wb")
    serialization_module.dump(obj, f, protocol=protocol, *args)
    f.close()

# Reading serial object
def read_object(path:str, compression="infer", serialization_module=pickle):
    path = format_path(path, str)

    if compression == "infer":
        _ , ext = os.path.splitext(path)
        if (ext == ".pkl") or (ext == ".dill"):
            compression = None
        if ext in {".pgz", ".gz"}:
            compression = "gzip"
        if ext in {".pbz2", ".bz2"}:
            compression = "bz2"
    if compression is not None:
        if compression == "gzip":
            f = gzip.open(path, "rb")
        if compression == "bz2":
            f = bz2.open(path, "rb")
    else:
        f = open(path, "rb")
    obj = serialization_module.load(f)
    f.close()
    return obj

# Importing a functions from a module
def read_script_as_module(name_module, path):
    # https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
    path = format_path(path, str)
    spec = importlib.util.spec_from_file_location(name_module, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time

# PyData
import pandas as pd

# Scikits
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from skbio.stats.ordination import pcoa
from ..io import write_object
from ..utils import is_query_class, to_precision, dict_filter, is_symmetrical
from ..symmetry import Symmetric
from ..visuals import plot_scatter

__all__ = ["MatrixDecomposition", "PrincipalComponentAnalysis","PrincipalCoordinatesAnalysis", "Manifold", "eigenprofiles_from_data"]
__all__ = sorted(__all__)


# General ordination methods
class CoreOrdinationMethods(object):
    # Get principal components
    def __getitem__(self, key):
        if type(key) == int:
            key = f"{prefix}{key}"
        return self.projection_[key]

    # Plotting
    def plot(self, precision=4, **args):
        """
        See soothsayer.visuals.plot_scatter for details on parameters.
        """
        xlabel = f"{self.prefix}1"
        ylabel = f"{self.prefix}2"
        if hasattr(self, "explained_variance_ratio_"):
            ev1 = to_precision(self.explained_variance_ratio_[f"{self.prefix}1"], precision=precision)
            ev2 = to_precision(self.explained_variance_ratio_[f"{self.prefix}2"], precision=precision)
            xlabel = f"{xlabel}({ev1})"
            ylabel = f"{ylabel}({ev2})"
        return plot_scatter(data=self.projection_, x=f"{self.prefix}1", y=f"{self.prefix}2", xlabel=xlabel, ylabel=ylabel, **args)

    # Plot Scree
    def plot_scree(self, title=None, color="teal", fig_kws=dict(), label_kws=dict(), title_kws=dict(), truncate=None, ax=None, mode="explained_variance", bar_kws=dict(), tick_kws=dict(), line_kws=dict(), style="seaborn-white"):
        _fig_kws = {"figsize":(8,3)}
        _fig_kws.update(fig_kws)
        _label_kws = {"fontsize":15}
        _label_kws.update(label_kws)
        _bar_kws = {"edgecolor":"black", "linewidth":1}
        _bar_kws.update(bar_kws)
        _line_kws = {"marker":"o",  "markeredgewidth":1, "markeredgecolor":"black"}
        _line_kws.update(line_kws)
        _tick_kws = {"fontsize":12}
        _tick_kws.update(tick_kws)
        _title_kws = {"fontsize":15, "fontweight":"bold"}
        _title_kws.update(title_kws)

        with plt.style.context(style):
            if ax is not None:
                fig = plt.gcf()
            else:
                fig, ax = plt.subplots(**_fig_kws)

            if mode == "explained_variance":
                self.explained_variance_ratio_[:truncate].plot(kind="bar",  color=color, ax=ax, **_bar_kws)
                ax.set_ylabel("Explained Variance", **_label_kws)
            if mode == "cumulative_variance":
                self.explained_variance_ratio_.cumsum()[:truncate].plot(kind="line", color=color, ax=ax, **_line_kws)
                ax.set_ylabel("Cumulative Explained Variance", **_label_kws)
            ax.set_xticklabels(ax.get_xticklabels(), **_tick_kws)
            if title is not None:
                ax.set_title(title, **_title_kws)
            return fig, ax

    # Save to file
    def to_file(self, path:str, compression="infer"):
        write_object(self, path, compression=compression)

# Sklearn Decomposition wrapper class
class MatrixDecomposition(object):
    """
    Wrapper for sklearn.decomposition objects
    https://scikit-learn.org/stable/modules/classes.html#module-sklearn.decomposition
    """
    def __init__(self, X:pd.DataFrame, center=True, scale=False, n_components=-1, base_model=PCA, model_kws=dict(), prefix="PC.", name=None, verbose=False):
        self.X = X.copy()
        self.X_input = self.X
        self.center = center
        self.scale = scale
        self.name = name
        self.prefix = prefix
        self.verbose = verbose
        self.n_components = self._check_n_components(*X.shape, n_components)

        # Preprocess
        if self.center == True:
            if verbose:
                print("Centering data", file=sys.stderr)
            self.X_input = self.X_input - self.X_input.mean(axis=0)
        if self.scale == True:
            if verbose:
                print("Scaling data", file=sys.stderr)
            self.X_input = self.X_input/self.X_input.std(axis=0)

        # Model
        self.model = self._build_model(base_model=base_model, X_input=self.X_input, model_kws=model_kws)

        # Explained Variance
        if hasattr(self.model, "explained_variance_ratio_"):
            if verbose:
                print("Adding`explained_variance_ratio_`", file=sys.stderr)
            self.explained_variance_ratio_ = pd.Series(self.model.explained_variance_ratio_, name="Explained variance ratio")
            self.explained_variance_ratio_.index = self.explained_variance_ratio_.index.map(lambda i: f"{prefix}{i+1}")

        # Components
        if hasattr(self.model, "components_"):
            if verbose:
                print("Adding`components_`", file=sys.stderr)
            self.components_ = self.model.components_

        # Project data onto eigenvectors
        if verbose:
            print("Adding`projection_`", file=sys.stderr)
        self.projection_ =  pd.DataFrame(
            self.model.transform(self.X_input),
            index = self.X_input.index,
        )
        self.projection_.columns = self.projection_.columns.map(lambda i: f"{prefix}{i+1}")

    # Check number of possible components
    def _check_n_components(self, n, m, n_components):
        possible_components = min(n, m)
        if n_components == -1:
            n_components = possible_components
        if (n_components > possible_components):
            if self.verbose:
                print(f"Data shape could not support {n_components} and is using the maximum {possible_components} instead.", file=sys.stderr)
            n_components = possible_components
        return n_components

    # Construct model
    def _build_model(self, base_model, X_input, model_kws):
        model = base_model(n_components=self.n_components, **model_kws)
        model.fit(X_input)
        return model

# Principal Component Analysis
class PrincipalComponentAnalysis(CoreOrdinationMethods):
    """
    See soothsayer.ordination.MatrixDecomposition for details on parameters.
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html#sklearn.decomposition.PCA
    """
    def __init__(self, X:pd.DataFrame, center=True, scale=False, n_components=-1, base_model=PCA, model_kws=dict(), prefix="PC.", name=None, verbose=False):

        self.core = MatrixDecomposition(
            X=X,
            center=center,
            scale=scale,
            n_components=n_components,
            base_model=base_model,
            model_kws=model_kws,
            prefix=prefix,
            name=name,
            verbose=verbose
        )
        self.__dict__.update(self.core.__dict__)


# Principal Coordinant Analysis
class PrincipalCoordinatesAnalysis(CoreOrdinationMethods):
    """
    Wrapper for skbio.stats.ordination.pcoa
    http://scikit-bio.org/docs/latest/generated/generated/skbio.stats.ordination.pcoa.html#skbio.stats.ordination.pcoa
    """
    def __init__(self, kernel, metric_type=None, node_type=None, prefix="PCoA.", name=None, verbose=False, symmetric_kws=dict()):
        if is_query_class(kernel, "DataFrame"):
            kernel = Symmetric(kernel, name=name, metric_type=metric_type, mode="dissimilarity", **symmetric_kws)
        self.kernel = kernel
        self.labels = kernel.labels
        self.node_type = node_type
        self.metric_type = metric_type
        self.name = name
        self.prefix = prefix
        self.verbose = verbose

        # Model
        self.model = self._build_model(kernel=kernel, base_model=pcoa)

        # Explained Variance
        if hasattr(self.model, "proportion_explained"):
            if verbose:
                print("Adding`explained_variance_ratio_`", file=sys.stderr)
            self.explained_variance_ratio_ = pd.Series(self.model.proportion_explained.values, name="Explained variance ratio")
            self.explained_variance_ratio_.index = self.explained_variance_ratio_.index.map(lambda i: f"{prefix}{i+1}")

        # Components
        if hasattr(self.model, "eigvals"):
            if verbose:
                print("Adding`eigvals_`", file=sys.stderr)
            self.eigvals_ = pd.Series(self.model.eigvals.values, name="Eigenvalues")
            self.eigvals_.index = self.eigvals_.index.map(lambda i: f"{prefix}{i+1}")


        # Project data onto eigenvectors
        if verbose:
            print("Adding`projection_`", file=sys.stderr)
        self.projection_ =  pd.DataFrame(
            self.model.samples.values,
            index = self.labels,
        )
        self.projection_.columns = self.projection_.columns.map(lambda i: f"{prefix}{i+1}")

    # Construct model
    def _build_model(self, kernel, base_model):
        df_dism = kernel.to_dense()
        model = base_model(df_dism)
        return model

# Manifold Learning
class Manifold(CoreOrdinationMethods):
    """
    Wrapper for sklearn.manifold objects
    https://scikit-learn.org/stable/modules/classes.html#module-sklearn.manifold
    """
    def __init__(self, X:pd.DataFrame, base_model=TSNE, n_components=2, compute_pca="infer", n_components__pca=50, pca_kws=dict(center=True, scale=False), random_state=0, prefix="infer", name=None, verbose=False, **model_kws):
        # Init
        base_model_name = str(base_model)[:-2].split(".")[-1]
        self.name = name
        if prefix == "infer":
            if base_model_name == "TSNE":
                prefix = f"t-SNE."
            else:
                prefix = f"{base_model_name}."
        self.prefix = prefix
        self.verbose = verbose

        # Is X precomputed?
        self.X = X.copy()
        self.X_input = self.X
        self.precomputed = self._check_precomputed(self.X, model_kws)
        if self.precomputed:
            compatible_for_precomputed = any(x == base_model_name for x in ["TSNE",  "MDS", "SpectralEmbedding", "spectral_embedding", "UMAP"])
            assert compatible_for_precomputed, f"{base_model_name} is not compatible for precomputed option"
            if base_model_name == "TSNE":
                model_kws["metric"] = "precomputed"
            if base_model_name == "MDS":
                model_kws["dissimilarity"] = "precomputed"
            if base_model_name == "SpectralEmbedding":
                model_kws["affinity"] = "precomputed"
            if base_model_name == "spectral_embedding":
                model_kws["adjacency"] = "precomputed"
        else:
            # Should PCA be computed?
            if compute_pca == "infer":
                if base_model_name == "TSNE":
                    compute_pca = True
            if compute_pca:
                # Check how many PCA components can be computed from the data
                self.n_components__pca = self._check_n_components(*self.X.shape, n_components=n_components__pca)
                if verbose:
                    print(f"Computing PCA for input data using {self.n_components__pca} components", file=sys.stderr)
                self.X_input = PrincipalComponentAnalysis(X=self.X, base_model=PCA, n_components=self.n_components__pca, **pca_kws).projection_

        # Random state
        self.random_state = random_state
        compatible_for_rs = ["LocallyLinearEmbedding", "TSNE", "MDS", "SpectralEmbedding", "locally_linear_embedding", "smacof", "spectral_embedding", "UMAP"]
        if base_model_name in compatible_for_rs:
            model_kws["random_state"] = self.random_state

        # Check number of components
        self.n_components = self._check_n_components(*self.X_input.shape, n_components)


         # Model
        self.model = self._build_model(base_model=base_model, X_input=self.X_input, model_kws=model_kws)
        self.__dict__.update(dict_filter(self.model.__dict__, keys=lambda x:x.endswith("_")))

        # Get projections/embeddings
        if hasattr(self, "embedding_"):
            self.projection_ =  pd.DataFrame(
            self.embedding_,
            index = self.X_input.index,
        )
        self.projection_.columns = self.projection_.columns.map(lambda i: f"{prefix}{i+1}")


    # Check number of possible components
    def _check_n_components(self, n, m, n_components):

        possible_components = min(n, m)
        if n_components == -1:
            n_components = possible_components
        if (n_components > possible_components):
            if self.verbose:
                print(f"Data shape could not support {n_components} and is using the maximum {possible_components} instead.", file=sys.stderr)
            n_components = possible_components
        return n_components

    # Check if data is precomputed
    def _check_precomputed(self, X, model_kws):
        # Preprocess
        precomputed = None
        if "metric" in model_kws:
            if model_kws["metric"] == "precomputed":
                precomputed = True
            else:
                precomputed = False
        if precomputed is None:
            if is_query_class(X, "Symmetric"):
                X = X.to_dense()
                precomputed = True
            else:
                precomputed = is_symmetrical(X, tol=1e-5)
        if self.verbose:
            print(f"Precomputed: {precomputed}", file=sys.stderr)
        return precomputed

    # Construct model
    def _build_model(self, base_model, X_input, model_kws):
        model = base_model(n_components=self.n_components, **model_kws)
        model.fit(X_input)
        return model

    # Make compatible with scree plotting
    def plot_scree(self, **args):
        print("Cannot create scree plot for Manifolds", file=sys.stderr)
        return self

# =======================
# Wraps on wraps on wraps
# =======================
# Get eigenprofiles from data
def eigenprofiles_from_data(X:pd.DataFrame, y:pd.Series, class_type:str = None, mode="eigenprofiles", pca_kws=dict(), verbose:bool=True, axis=0, prefix="PC."):
    """
    Calculate eigenprofiles from data
    2018-July-13
    """
    accepted_modes = ["eigenprofiles", "eigenvalues", None]
    assert mode in accepted_modes, f"'{mode}' is not in accepted modes: {accepted_modes}"
    if axis == 1:
        X = X.T.copy()
    idx_attributes = X.columns
    profiles = list()
    d_eigenvalues = dict()

    if mode is None:
        tasks = ["eigenprofiles", "eigenvalues"]
    else:
        tasks = [mode]
    for group, _X in X.groupby(y, axis=0):
        _X_T = _X.T
        pca_model = PrincipalComponentAnalysis(_X_T, n_components=1, prefix=prefix, **pca_kws)
        if "eigenprofiles" in tasks:
            pc1 = pca_model.projection_[f"{prefix}1"]
            eigenprofile = pd.Series(pc1, index=idx_attributes, name=group)
            profiles.append(eigenprofile)
        if "eigenvalues" in tasks:
            eigenvalue = pca_model.explained_variance_ratio_[f"{prefix}1"]
            d_eigenvalues[group] = eigenvalue
    output = list()
    if "eigenprofiles" in tasks:
        df_eigenprofiles = pd.DataFrame(profiles)
        df_eigenprofiles.index.name = class_type
        output.append(df_eigenprofiles)
    if "eigenvalues" in tasks:
        eigenvalues = pd.Series(d_eigenvalues, name=class_type)
        output.append(eigenvalues)
    assert len(output) > 0, "Check `X`, `y`, and `mode` because no computation was done"
    if len(output) == 1:
        return output[0]
    else:
        return tuple(output)



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys, time, multiprocessing
# ==============================================================================
# R Imports
# ==============================================================================
from rpy2 import robjects, rinterface
from rpy2.robjects.packages import importr
try:
    from rpy2.rinterface import RRuntimeError
except ImportError:
    from rpy2.rinterface_lib.embedded import RRuntimeError
from rpy2.robjects import pandas2ri
pandas2ri.activate()
R = robjects.r
NULL = robjects.rinterface.NULL
rinterface.set_writeconsole_regular(None)
# ==============================================================================
# All
# ==============================================================================



# Load R Packages
R_packages = dict()
def R_package_retrieve(package):
    """
    ["dynamicTreeCut", "WGCNA", "fastcluster", "phyloseq", "philr", "ape", "metagenomeSeq", "edgeR"]
    """
    try:
        return R_packages[package]
    except KeyError:
        try:
            R_packages[package] = importr(package)
            if package == "WGCNA":
                R_packages[package].enableWGCNAThreads(multiprocessing.cpu_count())
            return R_packages[package]
        except RRuntimeError:
            print(f"{package} is not available", file=sys.stderr)










#



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys
from collections import defaultdict, OrderedDict
import pandas as pd
import numpy as np
from statsmodels.sandbox.stats.multicomp import multipletests
# Soothsayer
from ..utils import is_dict

# Adjust p-values
def p_adjust(p_values:pd.Series, method="fdr", name=None, **kwargs):
    """
    Multiple test correction to adjust p-values

    fdr -> fdr_bh # https://www.statsmodels.org/stable/generated/statsmodels.stats.multitest.multipletests.html

    `bonferroni` : one-step correction
    `sidak` : one-step correction
    `holm-sidak` : step down method using Sidak adjustments
    `holm` : step-down method using Bonferroni adjustments
    `simes-hochberg` : step-up method  (independent)
    `hommel` : closed method based on Simes tests (non-negative)
    `fdr_bh` : Benjamini/Hochberg  (non-negative)
    `fdr_by` : Benjamini/Yekutieli (negative)
    `fdr_tsbh` : two stage fdr correction (non-negative)
    `fdr_tsbky` : two stage fdr correction (non-negative)

    """
    # Check type of input data
    index = None
    if is_dict(p_values):
        p_values = pd.Series(p_values)
    if isinstance(p_values, pd.Series):
        index = p_values.index
        p_values = p_values.values
    if method == "fdr":
        method = "fdr_bh"

    if index is not None:
        mask_null = np.isnan(p_values)
        if np.any(mask_null):
            scaffold = np.ones_like(p_values)*np.nan
            scaffold[np.logical_not(mask_null)] = multipletests(p_values[np.logical_not(mask_null)], method=method, **kwargs)[1]
            return pd.Series(scaffold, index=index)
        else:
            adjusted = multipletests(p_values, method=method, **kwargs)[1]
            return pd.Series(adjusted, index=index, name=name)
    else:
        num_nan = np.isnan(p_values).sum()
        assert  num_nan == 0, "Please remove the {} missing values".format(num_nan)
        return multipletests(p_values, method=method, **kwargs)[1]






#




# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, multiprocessing, itertools, copy, datetime, warnings
from collections import OrderedDict, defaultdict

# PyData
import pandas as pd
import numpy as np
import xarray as xr
import networkx as nx

# Biology
from Bio import SeqIO, Seq
import ete3
import skbio

# SciPy
from scipy import stats
from scipy.spatial import distance
try:
    from fastcluster import linkage
except ImportError:
    from scipy.cluster.hierarchy import linkage
    print("Could not import `linkage` from `fastcluster` and using `scipy.cluster.hierarchy.linkage` instead", file=sys.stderr)

# Statistics and Miscellaneous Machine Learning
from astropy.stats import biweight_midcorrelation, median_absolute_deviation
from sklearn.metrics.pairwise import pairwise_distances

# Soothsayer
from ..transmute.conversion import linkage_to_newick, name_ete_nodes
from ..r_wrappers.packages.WGCNA import bicor
from ..utils import is_symmetrical, force_symmetry

__all__ = ["Symmetric", "pairwise", "pairwise_tree_distance", "pairwise_logfc", "dense_to_condensed"]
__all__ = sorted(__all__)


# =======================================================
# Pairwise calculations
# =======================================================
def dense_to_condensed(X, name=None, assert_symmetry=True, tol=None):
    if assert_symmetry:
        assert is_symmetrical(X, tol=tol), "`X` is not symmetric with tol=`{}`".format(tol)
    labels = X.index
    index=pd.Index([*map(frozenset,itertools.combinations(labels, 2))], name=name)
    data = distance.squareform(X, checks=False)
    return pd.Series(data, index=index, name=name)

# Symmetrical dataframes represented as augment pd.Series
class Symmetric(object):
    """
    An indexable symmetric matrix stored as the lower triangle for space

    devel
    =====
    2018-August-16
    * Added __add__, __sub__, etc.
    * Removed conversion to dissimilarity for tree construction
    * Added .iteritems method

    Future:
    Take in a Symmetric object and pd.Series with a frozenset index
    Fix the diagonal arithmetic
    Replace self._dense_to_condensed to dense_to_condensed
    """
    def __init__(self, X:pd.DataFrame, data_type=None, metric_type=None, func_metric=None, name=None, mode="infer", metadata=dict(), force_the_symmetry=True):
        acceptable_modes = ["similarity", "dissimilarity", "statistical_test", "infer"]
        assert mode in acceptable_modes, f"`mode` must be in {acceptable_modes}"
        if mode == "infer":
            mode = self._infer_mode(X)
        assert np.any(X.isnull()) == False, "Check what is causing NaN and remove them."
        if force_the_symmetry:
            X = force_symmetry(X)
        else:
            assert is_symmetrical(X, tol=1e-10), "X is not symmetric"

        self.mode = mode
        self.data_type = data_type
        self.func_metric = func_metric
        if func_metric is not None:
            assert hasattr(func_metric, "__call__"), "`func_metric` isn't a function"
            if metric_type is None:
                metric_type = func_metric.__name__
        self.metric_type = metric_type
        self.name = name
        self.labels = X.index
        self.diagonal = np.copy(np.diagonal(X))

        diagonal_elements = set(self.diagonal)
        warnings.warn('If operations are applied to the Symmetric object they will not be represented by the diagonal when converting back to dense form. Fix this in future version.')
        self.data = self._dense_to_condensed(X)
        self.__synthesized__ = datetime.datetime.utcnow()
        self.num_nodes = X.shape[0]
        # Metadata
        self.metadata = dict([
            ("name",name),
            ("data_type",data_type),
            ("synthesized",self.__synthesized__.strftime("%Y-%m-%d %H:%M:%S")),
            ("num_nodes",self.num_nodes),
            ("mode", mode),
            ("metric_type", metric_type),
        ]
        )
        self.metadata.update(metadata)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.metadata)[1:-1]})"

    # ==
    # QC
    # ==
    def _infer_mode(self, X):
        diagonal = np.diagonal(X)
        diagonal_elements = set(diagonal)
        assert len(diagonal_elements) == 1, "Cannot infer relationships from diagonal because multiple values"
        assert diagonal_elements <= {0,1}, "Diagonal should be either 0.0 for dissimilarity or 1.0 for similarity"
        return {0.0:"dissimilarity", 1.0:"similarity"}[list(diagonal_elements)[0]]

    # =======
    # Built-in
    # =======
    def __getitem__(self, key):
        assert type(key)  not in [int,float,str], "`key` must be a non-string iterable"
        assert len(key) >= 2, "`key` must have at least 2 identifiers. e.g. ('A','B')"
        key = frozenset(key)
        if len(key) == 1:
            return {"similarity":1.0, "dissimilarity":0.0, "statistical_test":0.0}[self.mode]
        else:
            if len(key) > 2:
                key = [*map(frozenset, itertools.combinations(key, r=2))]
            return self.data[key]
    def __len__(self):
        return self.num_nodes
    def __iter__(self):
        for v in self.data:
            yield v
    def iteritems(self):
        return self.data.iteritems()

    # Maths
    def __add__(self, x):
        symmetric_clone = copy.deepcopy(self)
        symmetric_clone.data += x
        # self.diagonal += x
        return symmetric_clone
    def __radd__(self, x):
        self.diagonal += x
        return Symmetric.__add__(self,x)
    def __sub__(self, x):
        symmetric_clone = copy.deepcopy(self)
        symmetric_clone.data -= x
        return symmetric_clone
    def __rsub__(self, x):
        symmetric_clone = copy.deepcopy(self)
        symmetric_clone.data = x - symmetric_clone.data
        return symmetric_clone
    def __mul__(self, x):
        symmetric_clone = copy.deepcopy(self)
        symmetric_clone.data *= x
        return symmetric_clone
    def __rmul__(self, x):
        return Symmetric.__mul__(self,x)
    def __truediv__(self, x):
        symmetric_clone = copy.deepcopy(self)
        symmetric_clone.data /= x
        return symmetric_clone
    def __rtruediv__(self, x):
        symmetric_clone = copy.deepcopy(self)
        symmetric_clone.data = x / symmetric_clone.data
        return symmetric_clone
    def __floordiv__(self, x):
        symmetric_clone = copy.deepcopy(self)
        symmetric_clone.data //= x
        return symmetric_clone
    def __rfloordiv__(self, x):
        symmetric_clone = copy.deepcopy(self)
        symmetric_clone.data = x // symmetric_clone.data
        return symmetric_clone

    # =======
    # Utility
    # =======
    def set_diagonal(self, values):
        self.diagonal = values
        return self

    # ==========
    # Conversion
    # ==========
    def to_dense(self, subset=None, diagonal=None):
        data_dense = distance.squareform(self.data.values)
        if diagonal is None:
            if diagonal == "infer":
                diagonal = {"similarity":1.0, "dissimilarity":0.0, "statistical_test":0.0}[self.mode]
            else:
                diagonal = self.diagonal
        np.fill_diagonal(data_dense, diagonal)
        df_dense = pd.DataFrame(data_dense, index=self.labels, columns=self.labels)
        if subset is None:
            return df_dense
        else:
            # There should be a better way to do this
            return df_dense.loc[subset,subset]

    def to_condensed(self):
        return self.data

    def _dense_to_condensed(self, X):
        np.fill_diagonal(X.values, 0)
        index=pd.Index([*map(frozenset,itertools.combinations(self.labels, 2))], name=self.name)
        data = distance.squareform(X, checks=False)
        return pd.Series(data, index=index, name=self.name)

    def _condensed_to_dense(self, y):
        y.index = pd.MultiIndex.from_tuples(y.index, names=[None,None])
        X = y.unstack().loc[self.labels,self.labels]
        mask_null = X.isnull().values
        X.values[mask_null] = X.T.values[mask_null]
        return X

    def as_tree(self, method="ward", into=ete3.Tree, node_prefix="y"):
        assert self.mode == "dissimilarity", "mode must be dissimilarity to construct tree"
        if not hasattr(self,"Z"):
            self.Z = linkage(self.data.values,method=method)
        if not hasattr(self,"newick"):
            self.newick = linkage_to_newick(self.Z, self.labels)
        tree = into(newick=self.newick, name=self.name)
        return name_ete_nodes(tree, node_prefix)

    def as_graph(self, graph=None):
        if graph is None:
            graph = nx.Graph()
        graph.name = self.name
        for (node_A, node_B), weight in self.data.iteritems():
            graph.add_edge(node_A, node_B, weight=weight)
        return graph

    def copy(self):
        return copy.deepcopy(self)

# Pairwise interactions
def pairwise(X, metric="euclidean", axis=1, name=None, into=pd.DataFrame, mode="infer", signed=True, n_jobs=-1, check_metrics=True, symmetric_kws=dict()):
    """
    Compute pairwise interactions

    01-August-2018
    """

    def get_data_object(df_dense, into, mode, func_signed, metric, symmetric_kws, CORRELATION_METRICS, DISTANCE_METRICS, name, check_metrics):
        if hasattr(metric, "__call__"):
            metric_name = metric.__name__
        else:
            metric_name = metric

        if mode == "similarity":
            if check_metrics:
                assert metric_name not in DISTANCE_METRICS, f"Cannot compute simlilarity for {metric.__name__}"

        if mode == "dissimilarity":
            if metric_name in CORRELATION_METRICS:
                df_dense = 1 - func_signed(df_dense)
        if into in [Symmetric, pd.Series]:
            kernel = Symmetric(df_dense, mode=mode, name=name, **symmetric_kws)
            if into == Symmetric:
                return kernel
            if into == pd.Series:
                return kernel.data
        else:
            return df_dense


    fill_diagonal="auto"
    CORRELATION_METRICS = {"kendalltau", "pearson", "spearman", 'biweight_midcorrelation', 'bicor'}
    STATISTICAL_TEST_METRICS = {'ks_2samp', 'ttest_1samp', 'ttest_ind', 'ttest_ind_from_stats', 'ttest_rel', "wilcoxon","ranksums", "mannwhitneyu"}
    DISTANCE_METRICS = distance.__dict__["__all__"]

    # Assertions
    accepted_modes = ["similarity", "dissimilarity", "statistical_test", "infer"]
    assert mode in accepted_modes, f"{mode} is not in {accepted_modes}"

    # Metric name
    if hasattr(metric,"__call__"):
        metric_name = metric.__name__
    else:
        metric_name = metric
    # Infer mode
    if mode == "infer":
        if metric_name in CORRELATION_METRICS:
            mode = "similarity"
        if metric_name in STATISTICAL_TEST_METRICS:
            mode = "statistical_test"
        if metric_name in DISTANCE_METRICS:
            mode = "dissimilarity"
        assert mode != "infer", f"Unable to infer mode from metric = {metric}"
        print(f"Inferred mode as `{mode}`", file=sys.stderr)

    # Similarity transformation function
    functions = {
        True: lambda X: (X+1)/2,
        False: lambda X:np.abs(X),
        None: lambda X:X,
    }
    func_signed = functions[signed]

    # Axis
    if axis == 0: X_copy = X.copy()
    if axis == 1: X_copy = X.copy().T

    # Locals
    Ar_X = X_copy.values
    labels = X_copy.index

    # Placeholder
    df_dense = None

    # Pairwise interactions
    while df_dense is None:
        # Statistical test
        if metric_name in STATISTICAL_TEST_METRICS:
            metric = lambda u,v: getattr(stats, metric_name)(u,v)[1]
            df_dense = pd.DataFrame(distance.squareform(distance.pdist(Ar_X, metric=metric)), index=labels, columns=labels)
            fill_diagonal = 0.0
            if mode != "statistical_test":
                print(f"Convert `mode` from `{mode}` to `statistical_test`", file=sys.stderr)
                mode = "statistical_test"
            break
        # Euclidean distance
        if metric_name == "euclidean":
            df_dense = pd.DataFrame(distance.squareform(distance.pdist(Ar_X, metric="euclidean")), index=labels, columns=labels)
            fill_diagonal = 0.0
            break
        # Biweight midcorrelation
        if metric == "bicor":
            metric = biweight_midcorrelation
            try:
                df_dense = bicor(X_copy.T, condensed=False, self_interactions=True)
                break
            except:
                pass

        # Compute correlation and send the rest down the pipeline
        if metric_name in set(CORRELATION_METRICS) - set(["bicor", "biweight_midcorrelation"]):
            metric = metric_name
        if type(metric) == str:
            # Correlation
            if metric in CORRELATION_METRICS:
                df_dense = X_copy.T.corr(method = metric)
                break
            # Distance
            if metric_name in DISTANCE_METRICS:
                metric = getattr(distance, metric)

        # Check metric
        assert hasattr(metric, "__call__"), f"metric `{metric}` is unrecognized and should be converted into a callable at this point of the function"

        # Pairwise computation
        df_dense = pd.DataFrame(pairwise_distances(X_copy, metric=metric, n_jobs=n_jobs), index=labels, columns=labels)

        # Diagonal
        if fill_diagonal is not None:
            if fill_diagonal == "auto":
                fill_diagonal = {"similarity":1.0, "dissimilarity":0.0}[mode]
            fill_diagonal = np.repeat(fill_diagonal, X_copy.shape[0]).tolist()
            np.fill_diagonal(df_dense.values, fill_diagonal)

    # Check dense
    assert df_dense is not None, f"`df_dense` is None.  Check the following: metric={metric}"

    # Check NaN
    df_isnull = df_dense.isnull()
    if np.any(df_isnull):
        num_null = df_isnull.values.ravel().sum()
        warnings.warn(f"There are {num_null} null values")

    # Symmetric keywords
    _symmetric_kws={
        "metric_type":metric_name,
        "func_metric":metric if hasattr(metric, "__call__") else None,
        "metadata":{
            "input_shape":X.shape,
        },
    }
    _symmetric_kws.update(symmetric_kws)

    # Return object
    return get_data_object(df_dense, into, mode, func_signed, metric, _symmetric_kws, CORRELATION_METRICS, DISTANCE_METRICS, name, check_metrics)

# Pairwise distance on ete trees
def pairwise_tree_distance(tree, topology_only=True):
    d_topology = defaultdict(dict)
    leaves = tree.get_leaf_names()
    for i, id_obsv_A in enumerate(leaves):
        for j in range(i+1, len(leaves)):
            id_obsv_B = leaves[j]
            if id_obsv_A != id_obsv_B:
                d_topology[id_obsv_A][id_obsv_B] =  d_topology[id_obsv_B][id_obsv_A] = tree.get_distance(target=id_obsv_A, target2=id_obsv_B, topology_only=topology_only)
    df_dism_tree = pd.DataFrame(d_topology)
    np.fill_diagonal(df_dism_tree.values, 0)
    return df_dism_tree

# Pairwise log fold-change
def pairwise_logfc(X, idx_ctrl, idx_treatment, func_log=np.log2, name_ctrl="control", name_treatment="treatment"):
    """
    Positive values means they are greater in idx_treatment than in idx_ctrl.
    Negative values means they are greater in idx_ctrl than in idx_treatment.
    """
    # Init
    idx_attr = X.columns

    idx_ctrl = sorted(idx_ctrl)
    idx_treatment = sorted(idx_treatment)

    # Log Transform
    X_logscale = func_log(X)

    # Groups
    A_ctrl = X_logscale.loc[idx_ctrl,:].values
    A_treatment = X_logscale.loc[idx_treatment,:].values

    # Pairwise profiles
    logfc_profiles = np.vstack(A_treatment[:, np.newaxis] - A_ctrl)

    # Labels
    idx_pairwise_labels = itertools.product(idx_treatment, idx_ctrl)
    names = [name_treatment, name_ctrl]
    return pd.DataFrame(logfc_profiles, index=pd.MultiIndex.from_tuples(idx_pairwise_labels, names=names), columns=idx_attr)




# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, datetime, copy, warnings, pathlib
from collections import OrderedDict, defaultdict

# PyData
import pandas as pd
import numpy as np
import networkx as nx

# Biology
from Bio import SeqIO, Seq
import ete3
import skbio

# SciPy
from scipy.cluster import hierarchy as sp_hierarchy
try:
    from fastcluster import linkage
except ImportError:
    from scipy.cluster.hierarchy import linkage
    warnings.warn("Could not import `linkage` from `fastcluster` and using `scipy.cluster.hierarchy.linkage` instead")

## Matplotlib
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import rgb2hex
from mpl_toolkits.axes_grid1 import make_axes_locatable

# Soothsayer
from ..io import write_object
from ..r_wrappers.packages.dynamicTreeCut import cutree_dynamic
from ..utils import *
from ..ordination import PrincipalComponentAnalysis, eigenprofiles_from_data
from ..symmetry import Symmetric, pairwise
from ..transmute.conversion import linkage_to_newick, dism_to_linkage, ete_to_nx, name_ete_nodes
from ..networks import intramodular_connectivity
from ..tree import create_tree


# Hierarchical Clustering
def _func_dcoord(dcoord):
    return -np.log(dcoord+1)
class Agglomerative(object):
    # Sycamore, Ent, Cypress, Telperion, Bifurcation, Echelon, Yucca, Nimloth
    """
    # ==========
    # __future__
    # ==========
    * Incorporate missing values for track plots [BUG]
    * Add markers/annotations for specific nodes
    * Polar functionality for the following:
        * horizontal_lines
        * _plot_clusters_mpl
        * _plot_tracks_mpl
    # ==========
    # devel
    # ==========
    2019-May-29
    * Added spines to tracks
    * Added bar outline
    2019-March-11
    * Removed `cluster_mapping`
    * Added `add_secondary_classes` like `add_tracks`
    * Adapted `_plot_clusters_mpl` to plotting secondary classes as well
    * Changed `_plot_clusters_mpl` -> `_plot_classes_mpl`

    2019-February-22
    * Add label encoder
    * Clean up axes
    * Change `extract` to `get`
    * Add polar projection dendrogram
    * cdict and ndict -> *_colors and node_ampping
    * cluster_assignment -> cluster_mapping
    * Removed dbscan compatibility
    ___________________________________________________________________________________
    ___________________________________________________________________________________
    2018-August-16
    * Added ability to use Symmetric objects
    * Added as_graph method
    ___________________________________________________________________________________
    ___________________________________________________________________________________
    2018-August-04
    * Fixed case where node names have incompatible characters with newick strings
    ___________________________________________________________________________________
    # =====
    # Usage
    # =====
    X_modified = (X_iris + np.random.RandomState(0).normal(loc=10, size=(X_iris.shape)))
    df_dism = pairwise(X_modified, "euclidean", 0)

    # Instantiate the clustering object
    x = Agglomerative(df_dism, data=X_modified, name="iris", leaf_type="subject", tree_type=ete3.ClusterTree)
    # Hierarchical clustering using R's cuttreedynamic function (like WGCNA)
    x.cluster()
    # Add extra data to plot on dendrogram
    x.add_track("petal-length_1", X_iris["petal_length"], color=c_iris, plot_type="bar")
    x.add_track("petal-length_2", X_iris["petal_length"], color="black", plot_type="area")
    x.add_track("sepal-length", X_iris["sepal_length"], color=c_iris, plot_type="bar")
    # Plot the data
    x.plot(show_tracks=True, show_leaves=False, cluster_size=20, distance_label="euclidean distance", title="iris dataset w/ N(10,1) noise")
    print(x.silhouette(1))
    # ==========
    # Notes
    # ==========
    # Since ete3 trees are used and the node names can be limited using newick format, the leaf names are relabeled. In particular, check out _relabel and the linkage_to_newick function... it will make sense.
    # ==========
    # Known bugs
    # ==========
    *  Missing values causes track plotting to break
         * x.add_track("sepal-length-species_A", X_iris["sepal_length"][:50], color=c_iris[:50], plot_type="line", ylim=(X_iris["sepal_length"].min(),X_iris["sepal_length"].max()))
    * When using `ylim` with `add_track` the plot comes back as empty :/
    * Leaves must be strings for some reason...
    """
    def __init__(self,
                 kernel,
                 name=None,
                 data=None,
                 leaf_axis="infer",
                 method="ward",
                 leaf_type=None,
                 tree_type=ete3.ClusterTree,
                 palette="hls",
                 outlier_color="white",
                 metadata=dict(),
                 tree_kws=dict(),
                 polar_kws=dict(),
                 rectilinear_kws=dict(),
                ):
        # Check the kernel object type
        accepted_kernel_types = {pd.DataFrame, Symmetric}
        assert type(kernel) in accepted_kernel_types, f"`kernel` type must be one of the following: {accepted_kernel_types}"

        # pd.DataFrame -> Symmetric
        if isinstance(kernel, pd.DataFrame):
            kernel = Symmetric(X=kernel, data_type=leaf_type, name=name, mode="dissimilarity", metadata=metadata)

        # Base
        self.name = name
        self.data = data
        self.kernel = kernel

        # Clustering
        self.method = method
        self.linkage_labels = self.kernel.labels.copy()
        self.num_leaves = len(self.linkage_labels)
        self._relabel = dict(zip(range(self.num_leaves), self.linkage_labels))
        self.Z = linkage(self.kernel.data.values, method=method)
        if leaf_axis == "infer":
            if data is not None:
                axis_0_overlap = len(set(self.linkage_labels) & set(data.index))
                axis_1_overlap = len(set(self.linkage_labels) & set(data.columns))
                leaf_axis = np.argmax([axis_0_overlap, axis_1_overlap])
            else:
                leaf_axis = None
        self.leaf_axis = leaf_axis
        self._clusters_exist = False

        # Data
        self.tree_type = tree_type
        if tree_type == ete3.ClusterTree:
            if data is not None:
                if leaf_axis == 0:
                    matrix_string = dataframe_to_matrixstring(data)
                elif leaf_axis == 1:
                    matrix_string = dataframe_to_matrixstring(data.T)
            else:
                matrix_string = None
            _tree_kws = {"text_array":matrix_string}
        else:
            _tree_kws = {}
        _tree_kws.update(tree_kws)

        #Tree
        self.newick = linkage_to_newick(self.Z, range(self.num_leaves))
        self.tree = tree_type(newick=self.newick, **_tree_kws)
        self.tree.name = name # Hack for using ete3.ClusterTree objects b/c there's no name argument
        self._relabel_tree(self.tree, self._relabel)
        name_ete_nodes(self.tree, node_prefix="y")
        self.dendrogram = sp_hierarchy.dendrogram(self.Z, labels=self.linkage_labels, no_plot=True)

        # Colors
        self.palette = palette
        self.outlier_color = outlier_color

        # Polar

        self._polar_kws = {"pad":0.02, "n_smooth":100, "func_dcoord":_func_dcoord}
        self._polar_kws.update(polar_kws)

        # Rectilinear
        self._rectilinear_kws = {}
        self._rectilinear_kws.update(rectilinear_kws)

        # Leaves
        self.leaf_type = leaf_type
        self.leaves = pd.Index(self.dendrogram["ivl"], name=leaf_type)
        self.leaf_ticks_rectilinear = np.arange(5, self.num_leaves * 10 + 5, 10)
        t_grid = np.linspace(self._polar_kws["pad"]/2, 1-self._polar_kws["pad"]/2, self.num_leaves)
        self.leaf_ticks_polar = 2*np.pi*t_grid


        self.leaf_positions_rectilinear = dict(zip(self.leaves,self.leaf_ticks_rectilinear))
        self.leaf_positions_polar = dict(zip(self.leaves, self.leaf_ticks_polar))

        # Extension data
        self.tracks = OrderedDict()
        self.secondary_classes = OrderedDict()
        self.axes = list()

        # Metadata
        self.__synthesized__ = datetime.datetime.utcnow()
        self.metadata = {
            "name":self.name,
            "method":self.method,
            "num_leaves":self.num_leaves,
            "leaf_type":leaf_type,
            "tree_type":tree_type.__name__,
            "synthesized":self.__synthesized__.strftime("%Y-%m-%d %H:%M:%S"),
            "num_tracks":0}
        self.metadata.update(metadata)

        # Leaf attributes
        self.leaf_attributes = pd.DataFrame([
            pd.Series(self.leaf_positions_rectilinear, name="tick_position(rectilinear)"),
            pd.Series(self.leaf_positions_polar, name="tick_position(polar)"),
        ]).T.loc[self.leaves,:]
        self.leaf_attributes.index.name = self.leaf_type
    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}({str(self.metadata)[1:-1]})"

    def copy(self):
        return copy.deepcopy(self)



    # =======
    # Conversion
    # =======
    def _relabel_tree(self, tree, _relabel):
        for node in tree.traverse():
            if node.is_leaf():
                node.name = _relabel[int(node.name)]
        return tree

    # =======
    # Utility
    # =======
    def to_file(self, path, compression="infer"):
        write_object(self, path, compression=compression, serialization_module=pickle, protocol=pickle.HIGHEST_PROTOCOL)

    # ==========
    # Clustering
    # ==========
    def cluster(self, deep_split=2, min_c=3, cut_method="hybrid", cut_height=None, algo_kws=dict(), outlier_module=None):
        """
        Future: Cluster using the actual ete3 nodes
        """
        accepted_arguments = {"hybrid", "static"}
        assert cut_method in accepted_arguments, f"`cut_method` must be one of the following: {accepted_arguments}"

        # Utility
        def _process_subtrees(tree, clusters):
            d_cluster_node = dict(zip(clusters.keys(), [np.nan]*len(clusters)))
            for cluster, leaves in clusters.iteritems():
                subtree = tree.get_common_ancestor(*leaves)
                d_cluster_node[cluster] = subtree.name
            return d_cluster_node

        def _process_static_clusters(cluster_mapping, min_c, outlier_module):
            # Sort clusters by size
            sorted_clusters = sorted(pd_series_collapse(cluster_mapping, type_collection=pd.Index).values, key= len, reverse=True)
            # Group the outliers
            d_cluster_leaves = defaultdict(list)

            i = 0
            for cluster in sorted_clusters:
                # WIP
                if len(cluster) < min_c: # There will never be clusters smaller than this from cutreedynamic
                    d_cluster_leaves[outlier_module] += cluster.tolist()
                else:
                    d_cluster_leaves[i] = cluster.tolist()
                    i += 1
            return d_cluster_leaves

        cluster_names = np.nan
        outlier_module = -1
        # Static tree cutting
        if cut_method == "static":
            _algo_kws = {"criterion":"distance"}
            _algo_kws.update(algo_kws)
            # Cut tree at threshold and get clusters indices
            clusters = sp_hierarchy.fcluster(Z=self.Z, t=cut_height, **_algo_kws)
            cluster_mapping = pd.Series(clusters, index=self.linkage_labels)
            # Relabel and identify outliers
            d_cluster_leaves = _process_static_clusters(cluster_mapping, min_c=min_c, outlier_module=outlier_module)
            cluster_mapping = pd.Series(dict_expand(d_cluster_leaves, into=OrderedDict), name=self.name).sort_values()

        # Hybrid Dynamic Tree Cutting
        if cut_method == "hybrid":
            _algo_kws = {}
            _algo_kws.update(algo_kws)
            cluster_mapping = cutree_dynamic(self.kernel, cut_method=cut_method, method=self.method, minClusterSize=min_c, deepSplit=deep_split, name=self.name, **_algo_kws).astype(int) #Outlier module

        # Index the clusters by the leaves
        self.cluster_mapping = cluster_mapping[self.leaves]

        # Set the colors
        self.colors = Chromatic.from_classes(cluster_mapping, name=self.name, palette=self.palette, format="hex", outlier_color=self.outlier_color, outlier_value=outlier_module)
        self.leaf_colors = self.colors.obsv_colors
        self.cluster_colors = self.colors.class_colors


        # Global objects
        # ==============
        # Add colors and clusters
        self.leaf_attributes["cluster"] = self.cluster_mapping
        self.leaf_attributes["color"] = self.leaf_colors
        # Add nodes
        self.node_mapping = _process_subtrees(self.tree, self.get_clusters(grouped=True))

        # Update
        self.cut_method = cut_method
        if cut_method == "static":
            self.cut_height = cut_height
        self.outlier_module = outlier_module
        self.num_clusters = len(self.cluster_mapping[self.cluster_mapping.map(lambda x:x != outlier_module)].unique())
        self.clusters = sorted(self.get_clusters(grouped=True).keys())
        self.metadata["num_clusters"] = self.num_clusters
        self._clusters_exist = True
        self._outlier_module_exists = outlier_module in self.cluster_mapping.index

        # Cluster attributes
        self.cluster_attributes = pd.DataFrame([
            pd.Series(self.cluster_colors, name="color"),
            pd.Series(self.node_mapping, name="node"),
            pd.Series(self.get_clusters(grouped=True).map(set), name="leaves"),
        ]).T
        self.cluster_attributes["size"] = self.cluster_attributes["leaves"].map(len)
        self.cluster_attributes.index.name = "id_cluster"
        return self
    # ======
    # Groups
    # ======
    def add_secondary_class(self, name, mapping:pd.Series, class_colors=None, pad=0.1, percent_of_figure = 20, palette="hls", desat=None):
        mapping = pd.Series(mapping)
        missing_leaves = set(self.leaves) - set(mapping.index)
        assert len(missing_leaves) == 0, "`mapping` does have an secondary class for each leaf.  In particular, the following:\n{}".format(missing_leaves)
        self.leaf_attributes[name] = mapping
        if class_colors is None:
            class_colors = map_colors(mapping.unique(),mode=0, palette=palette, desat=desat)
        self.secondary_classes[name] = {"mapping":mapping, "class_colors":class_colors, "pad":pad, "size":f"{percent_of_figure}%"}
        return self

    # ======
    # Tracks
    # ======
    def add_track(self, name, data, alpha=0.618, color="black", pad=0.25, percent_of_figure = 20, ylim=None, plot_type="bar", spines=["top", "bottom", "left", "right"], fillna=0, plot_kws=dict(), label_kws=dict(),  bar_width="auto", check_missing_values=False, convert_to_hex=False, horizontal_kws=dict(), horizontal_lines=[]):
        """
        type(track) should be a pd.Series
        """
        # Keywords
        if plot_type == "bar":
            _plot_kws = {"linewidth":1.382, "alpha":alpha, "width":max(5,self.num_leaves/30) if bar_width == "auto" else bar_width}
        elif plot_type == "area":
            _plot_kws = {"linewidth":0.618, "alpha":0.5}
        else:
            _plot_kws = {}
        _plot_kws.update(plot_kws)
        _horizontal_kws = {"linewidth":1, "linestyle":"-", "color":"black"}
        _horizontal_kws.update(horizontal_kws)

        # Color of the track
        if color is None:
            color = "black"
        if isinstance(color,str):
            color = pd.Series([color]*len(self.leaves), index=self.leaves)
        if is_dict(color):
            color = pd.Series(color, index=index)
        if convert_to_hex:
            _sampleitem_ = color[0]
            assert mpl.colors.is_color_like(_sampleitem_), f"{_sampleitem_} is not color-like"
            if is_rgb_like(_sampleitem_):
                color = color.map(rgb2hex)
        # Check for missing values
        if check_missing_values:
            assert set(data.index) == set(self.leaves), "`data.index` is not the same set as `self.leaves`"
            assert set(color.index) == set(self.leaves), "`color.index` is not the same set as `self.leaves`"
        if plot_type in ["area","line"]:
            if color.nunique() > 1:
                print("Warning: `area` and `line` plots can only be one color.  Defaults to first color in iterable.", file=sys.stderr)
        # Horizontal lines
        if is_number(horizontal_lines):
            horizontal_lines = [horizontal_lines]
        # Convert to ticks
        data = data[self.leaves]
        color = color[self.leaves]
        # Store data
        self.tracks[name] = {"data":data, "color":color, "pad":pad, "size":f"{percent_of_figure}%", "plot_type":plot_type, "plot_kws":_plot_kws,  "ylim":ylim, "horizontal_lines":horizontal_lines, "horizontal_kws":_horizontal_kws, "spines":spines}
        self.metadata["num_tracks"] += 1

        return self

    # ============
    # get Data
    # ============
    def _get(self, query, field, grouped):
        try:
            if not grouped:
                if query is None:
                    return self.leaf_attributes[field]

                else:
                    if is_nonstring_iterable(query):
                        mask = self.leaf_attributes[field].map(lambda x: x in query)
                        idx_leaves_query = self.leaf_attributes[field][mask].index
                        idx_leaves_query.name = tuple(sorted(query))
                        return idx_leaves_query

                    else:
                        return self.leaf_attributes[field].compress(lambda x: x == query).index

            else:
                return pd_series_collapse(self.leaf_attributes[field], type_collection=pd.Index)

        except AttributeError:
            print("AttributeError: You must define or calculate the clusters first with `cluster` method", file=sys.stderr)
            return None

    # Clusters
    def get_clusters(self, query=None, grouped=False):
        return self._get(query=query, field="cluster", grouped=grouped)

    # Colors
    def get_colors(self, query=None, grouped=False):
        return self._get(query=query, field="color", grouped=grouped)

    # Sizes
    def get_sizes(self, query=None):
        if not is_nonstring_iterable(query):
            query = [query]
        return self.cluster_attributes.loc[slice(*query),"size"]

    # Nodes
    def get_nodes(self, query=None):
        if not is_nonstring_iterable(query):
            query = [query]
        return self.cluster_attributes.loc[slice(*query),"node"]

    # Subtrees
    def get_subtrees(self, query=None):
        def _search_tree(tree, query):
            node_name = self.node_mapping[query]
            return tree.search_nodes(name=node_name)[0]
        if query is not None:
            return _search_tree(self.tree, query)
        else:
            return pd.Series([*map(lambda x:_search_tree(self.tree, x), self.clusters)], index=pd.Index(self.clusters, name="subtrees"))
    # Tracks
    def get_tracks(self, query=None):
        if len(self.tracks) > 0:
            if query is not None:
                return self.tracks[query]
            else:
                df_tracks = pd.DataFrame(OrderedDict([(name, values["data"]) for name, values in self.tracks.items()]))
                df_tracks.index = self.leaves
                return df_tracks
    # ============
    # Plotting
    # ============
    def _plot_dendrogram_mpl(self, icoord, dcoord, xlim, ylim, color, ax, line_kws, n_smooth, distance_tick_kws, distance_label, show_distance, horizontal_lines, vertical_lines, func_dcoord):
        # Smoothing segments for polar plotting
        def _smooth_segment(seg, n_smooth):
            # https://stackoverflow.com/questions/51936574/how-to-plot-scipy-hierarchy-dendrogram-using-polar-coordinates
            # Note, this and the polar plotting has been adapted from above
            # Acknowledgements to @chthonicdaemon ^^^
            return np.concatenate([[seg[0]], np.linspace(seg[1], seg[2], n_smooth), [seg[3]]])

        # Plot dendrogram
        for xs, ys in zip(icoord, dcoord):
            if ax.name == "polar":
                xs = _smooth_segment(xs, n_smooth=n_smooth)
                ys = _smooth_segment(ys, n_smooth=n_smooth)
            ax.plot(xs, ys,  color, **line_kws)

        # Add vertical lines
        if bool(vertical_lines):
            if is_nonstring_iterable(vertical_lines) and not is_dict(vertical_lines):
                vertical_lines = {pos:dict() for pos in vertical_lines}
            for pos, kws in vertical_lines.items():
                ax.axvline(pos, **kws)
        # Add horizontal lines
        if bool(horizontal_lines):
            if ax.name == "polar":
                print("Warning: `horizontal_lines` visualization is currently not available with polar projection", file=sys.stderr)
            else:
                if is_nonstring_iterable(horizontal_lines) and not is_dict(horizontal_lines):
                    horizontal_lines = {pos:dict() for pos in horizontal_lines}
                for pos, kws in horizontal_lines.items():
                    ax.axhline(func_dcoord(pos), **kws)


        # Adjust ticks
        if ax.name == "polar":
            ax.set_rlabel_position(0)
            ax.set_xticks(self.leaf_ticks_polar)
        else:
            ax.set_ylim(ylim)
            ax.set_xlim(xlim)

            #Distance label
            conditions = dict()
            conditions[1] = distance_label is not None
            conditions[2] = show_distance
            if all(conditions.values()):
                ax.set_ylabel(distance_label, **distance_tick_kws)
        if not show_distance:
            ax.set_yticklabels([])
        ax.set_xticklabels([])

        return ax



    def _plot_track_mpl(self, name, divider, xlim, ylim, show_track_ticks, show_track_labels, label_kws):
        ax = divider.append_axes("bottom", size=self.tracks[name]["size"], pad=self.tracks[name]["pad"])
        data = self.tracks[name]["data"]
        c = self.tracks[name]["color"]
        kws = self.tracks[name]["plot_kws"]
        # Set leaf tick positions
        leaf_positions = self.leaf_positions_rectilinear #leaf_positions = {True:self.leaf_positions_polar, False:self.leaf_positions_rectilinear}[polar]
        data.index = data.index.map(lambda x:leaf_positions[x])
        c.index = c.index.map(lambda x:leaf_positions[x])

        if self.tracks[name]["plot_type"] == "bar":
            ax.bar(data.index, data.values, color=c, **kws)
            kws_outline = kws.copy()
            del kws_outline["alpha"]
            ax.bar(data.index, data.values, color="none", edgecolor=c, **kws_outline)
        else:
            if self.tracks[name]["plot_type"] == "area":
                data.plot(kind="area", color=c.values, ax=ax, **kws)
                data.plot(kind="line", color=c.values, alpha=1, ax=ax)
            else:
                data.plot(kind=self.tracks[name]["plot_type"], color=c.values, ax=ax,  **kws)
        if len(self.tracks[name]["horizontal_lines"]):
            for y_pos in self.tracks[name]["horizontal_lines"]:
                ax.axhline(y_pos, **self.tracks[name]["horizontal_kws"])
        # Spines
        for spine in ["top", "bottom", "left", "right"]:
            if spine not in self.tracks[name]["spines"]:
                ax.spines[spine].set_visible(False)
        ax.set_xlim(xlim)
        ax.set_xticklabels([])
        if self.tracks[name]["ylim"] is not None:
            ax.set_ylim(ylim)
        if not show_track_ticks:
            ax.set_yticklabels([])
        if show_track_labels:
            ax.set_ylabel(name, **label_kws, ha="right", va="center")
        return ax


    def _plot_classes_mpl(self, size, pad, divider, xlim, show_cluster_labels, fill_kws, cluster_kws, name, track_label, label_kws, show_track_labels):
        ax = divider.append_axes("bottom", size=size, pad=pad)
        mode = {True:"primary", False:"secondary"}[name is None]
        if mode == "primary":
            mapping = self.leaf_attributes["cluster"]
            class_colors = self.cluster_colors

            # Labels
            if show_cluster_labels:
                for cluster, leaves in self.get_clusters(grouped=True).iteritems():
                    idx_middleleaf = leaves[len(leaves)//2]
                    label_position = self.leaf_positions_rectilinear[idx_middleleaf]
                    label = cluster
                    ax.annotate(label, xy=(label_position, 0.5), **cluster_kws)
        elif mode == "secondary":
            mapping = self.secondary_classes[name]["mapping"]
            class_colors = self.secondary_classes[name]["class_colors"]

        for leaf, position in self.leaf_positions_rectilinear.items():
            color = class_colors[mapping[leaf]]
            # Cluster rectangle
            ax.fill_between(x=[position - 5,position + 5], y1=0, y2=1, color=color, **fill_kws)

        if all([show_track_labels, (track_label is not None)]):
            ax.set_ylabel(track_label, **label_kws, ha="right", va="center")

        ax.set_xlim(xlim)
        ax.set_yticklabels([])
        return ax


    # Plot controller
    def plot(self,
             branch_color="auto",
             leaf_colors=None,
             leaf_font_size=15,
             leaf_rotation="auto",
             style="seaborn-white",
             show_leaves=False,
             show_clusters=True,
             show_cluster_labels=True,
             show_secondary_classes=True,
             show_tracks=True,
             show_track_ticks=False,
             show_track_labels=True,
             show_spines=True,
             show_xgrid=False,
             show_ygrid=False,
             show_distance=True,
             cluster_size=20,
             cluster_pad=0.1,
             cluster_label="cluster",
             figsize=(21,5),
             title=None,
             ax=None,
             legend=False,
             fig_kws=dict(),
             legend_kws=dict(),
             label_kws=dict(),
             line_kws=dict(),
             fill_kws=dict(),
             cluster_kws=dict(),
             leaf_tick_kws=dict(),
             distance_tick_kws=dict(),
             distance_label="Distance",
             title_kws=dict(),
             polar=False,
             pad=None,
             horizontal_lines=dict(),
             vertical_lines=dict(),
            ):

        # Initialize
        self.axes = list()
        orientation = "top" # Does not work with bottom, left, or right yet 2018-June-26

        # Keywords
        _fig_kws = {"figsize":figsize}
        _fig_kws.update(fig_kws)
        _label_kws = {"fontsize":15, "rotation":0, "horizontalalignment":"right"}
        _label_kws.update(label_kws)
        _legend_kws = {'fontsize': 12, 'frameon': True, 'facecolor': 'white', 'edgecolor': 'black', 'loc': 'center left', 'bbox_to_anchor': (1, 0.5)}
        _legend_kws.update(legend_kws)
        _line_kws = {"linestyle":"-", "linewidth":1.382}
        _line_kws.update(line_kws)
        _fill_kws = {"alpha":0.618}
        _fill_kws.update(fill_kws)
        _cluster_kws = {"fontsize":12, "verticalalignment":"center", "horizontalalignment":"center", "bbox":dict(boxstyle="square", fc="w", pad=0.25, alpha=0.85)}
        _cluster_kws.update(cluster_kws)
        _distance_tick_kws = {"fontsize":15, "rotation":{True:0, False:90}[all([show_tracks, self.metadata["num_tracks"] > 0])]}
        if _distance_tick_kws["rotation"] == 0:
            _distance_tick_kws["horizontalalignment"] = "right"
        _distance_tick_kws.update(distance_tick_kws)
        if leaf_rotation in ["auto",None]:
            leaf_rotation = {True:90, False:0}[orientation in ["top","bottom"]]
        else:
            leaf_rotation = leaf_rotation
        _leaf_tick_kws = {"fontsize":12, "rotation":leaf_rotation}
        _leaf_tick_kws.update(leaf_tick_kws)
        _title_kws = {"fontsize":18, "fontweight":"bold","y":0.95}
        _title_kws.update(title_kws)


        # Branch colors
        if branch_color == "auto":
            branch_color = {True:"white", False:"black"}[style == "dark_background"]

        # Legend
        if legend == False:
            legend = None

        if legend is not None:
            if legend == True:
                if hasattr(self, "cluster_colors"):
                    legend = self.cluster_colors
            else:
                if isinstance(legend, pd.Series):
                    legend = legend.to_dict(OrderedDict)
        # Adjust coordinates for polar
        icoord=np.asarray(self.dendrogram["icoord"], dtype=float)
        dcoord=np.asarray(self.dendrogram["dcoord"], dtype=float)

        func_dcoord = lambda dcoord: dcoord

        if polar:
            func_dcoord = self._polar_kws["func_dcoord"]
            dcoord = func_dcoord(dcoord)
            icoord_max = icoord.max()
            icoord_min = icoord.min()
            icoord = ((icoord - icoord_min)/(icoord_max - icoord_min) * (1 - self._polar_kws["pad"]) + self._polar_kws["pad"]/2)*2*np.pi
        else:
            if "func_dcoord" in self._rectilinear_kws:
                func_dcoord = self._rectilinear_kws["func_dcoord"]
                dcoord = func_dcoord(dcoord)

        # Limits
        # Independent variable plot width
        tree_width = len(self.leaves) * 10
        xlim = (0,tree_width)
            # Depenendent variable plot height
        maximum_height = dcoord.flatten().max() #max(self.Z[:, 2])
        tree_height = maximum_height + maximum_height * 0.05
        ylim = (0,tree_height)

        # Plotting
        with plt.style.context(style):
            if ax is None:
                fig = plt.figure(**_fig_kws)
                ax_dendrogram = fig.add_subplot(111, polar=polar)
            else:
                fig = plt.gcf()
                ax_dendrogram = ax

            # Arguments for dendrogram plotting
            args_dendro = dict(
                icoord=icoord,
                dcoord=dcoord,
                color=branch_color,
                xlim=xlim,
                ylim=ylim,
                ax=ax_dendrogram,
                line_kws=_line_kws,
                distance_tick_kws=_distance_tick_kws,
                distance_label=distance_label,
                n_smooth=self._polar_kws["n_smooth"],
                show_distance=show_distance,
                horizontal_lines=horizontal_lines,
                vertical_lines=vertical_lines,
                func_dcoord=func_dcoord,
            )
            ax_dendrogram = self._plot_dendrogram_mpl(**args_dendro)
            self.axes.append(ax_dendrogram)
            # Add tracks
            divider = make_axes_locatable(ax_dendrogram)
            if show_tracks:
                if not polar:
                    if len(self.tracks) > 0:

                        for name in self.tracks:
                            if "edgecolor" in self.tracks[name]["plot_kws"]:
                                if self.tracks[name]["plot_kws"]["edgecolor"] == "auto":
                                    self.tracks[name]["plot_kws"]["edgecolor"] = {True:"white", False:"black"}[style == "dark_background"]
                            args_track = dict(
                                name=name,
                                divider=divider,
                                xlim=xlim,
                                ylim=ylim,
                                show_track_ticks=show_track_ticks,
                                show_track_labels=show_track_labels,
                                label_kws=_label_kws,
                            )
                            ax_track = self._plot_track_mpl( **args_track)
                            self.axes.append(ax_track)
                else:
                    print("Warning: `show_tracks` visualization is currently not available with polar projection", file=sys.stderr)
            # Add secondary classes
            if all([bool(show_secondary_classes), bool(self.secondary_classes)]):
                if not polar:
                    if show_secondary_classes == True:
                        show_secondary_classes = list(self.secondary_classes.keys())
                    assert set(show_secondary_classes) <= set(self.secondary_classes), "`show_secondary_classes` must be either `True`, `False`, or a subset of secondary classes that have been added manually"
                    for name, data in self.secondary_classes.items():
                        args_secondary_class = dict(
                            size=data["size"],
                            pad=data["pad"],
                            divider=divider,
                            xlim=xlim,
                            show_cluster_labels=False,
                            fill_kws=_fill_kws,
                            cluster_kws=_cluster_kws,
                            name=name,
                            track_label=name,
                            label_kws=_label_kws,
                            show_track_labels=show_track_labels,

                        )
                        ax_secondary_class = self._plot_classes_mpl( **args_secondary_class)
                        self.axes.append(ax_secondary_class)
                else:
                    print("Warning: `show_secondary_classes` visualization is currently not available with polar projection", file=sys.stderr)

            # Add clusters
            if all([show_clusters, self._clusters_exist]):
                if not polar:
                    args_clusters = dict(
                        size=f"{cluster_size}%",
                        pad=cluster_pad,
                        divider=divider,
                        xlim=xlim,
                        show_cluster_labels=show_cluster_labels,
                        fill_kws=_fill_kws,
                        cluster_kws=_cluster_kws,
                        name=None,
                        track_label=cluster_label,
                        label_kws=_label_kws,
                        show_track_labels=show_track_labels,
                    )
                    ax_clusters = self._plot_classes_mpl( **args_clusters)
                    self.axes.append(ax_clusters)
                else:
                    print("Warning: `show_clusters` visualization is currently not available with polar projection", file=sys.stderr)



            # Leaves
            if show_leaves:
                if leaf_colors is None:
                    if style == "dark_background":
                        leaf_colors = "white"
                    else:
                        leaf_colors = "black"
                if mpl.colors.is_color_like(leaf_colors):
                    leaf_colors = pd.Series([leaf_colors]*self.num_leaves, index=self.leaves)

                if len(self.axes) > 1:
                    [*map(lambda ax_query: ax_query.get_xaxis().set_visible(False), self.axes[:-1])]
                # Add leaf tick labels to polar projection plot
                if polar:
                    rmax = ax_dendrogram.get_rmax()
                    # Set pad
                    if pad is None:
                        rmin = ax_dendrogram.get_rmin()
                        pad = 0.01*rmin
                    # Add labels
                    for i, label in enumerate(self.axes[-1].get_xticklabels()):
                        leaf = self.leaves[i]
                        t = self.leaf_positions_polar[leaf]
                        label = self.axes[-1].text(t, rmax+pad, leaf, color=leaf_colors[leaf], transform=label.get_transform(), ha="center", va="center")
                        label.set_rotation(np.rad2deg(t))
                # Add leaf tick labels to rectilinear projection plot
                else:
                    self.axes[-1].set_xticks(self.leaf_ticks_rectilinear)
                    self.axes[-1].set_xticklabels(self.leaves, **_leaf_tick_kws)
                    self.axes[-1].xaxis.set_ticks_position("bottom")
                    for label in self.axes[-1].get_xticklabels():
                        label.set_color(leaf_colors[label.get_text()])
            else:
                self.axes[-1].set_xticklabels([])

            # Spines
            if polar:
                ax_dendrogram.spines["polar"].set_visible(show_spines)
            else:
                for spine_type, spine in ax_dendrogram.spines.items():
                    spine.set_visible(show_spines)

            # Legend
            if legend is not None:
                ax_dendrogram.legend(*format_mpl_legend_handles(legend), **_legend_kws)

            # Manage grids

            [*map(lambda ax_query: ax_query.xaxis.grid(show_xgrid), self.axes)]
            [*map(lambda ax_query: ax_query.yaxis.grid(show_ygrid), self.axes)]


            # Title
            if title is not None:
                fig.suptitle(title, **_title_kws)
            return fig, self.axes


    # ============
    # Analysis
    # ============
    def silhouette(self, cluster, func_distance=None, into=pd.Series):
        if self.tree_type == ete3.ClusterTree:
            subtree = self.get_subtrees(cluster)
            scores = subtree.get_silhouette(fdist=func_distance)
            labels = ["silhouette", "intracluster_dist", "intercluster_dist"]
            output = pd.Series(scores, index=labels, name=f"cluster_{cluster}")
            return into(output)
        else:
            print("Warning: `silhouette` score can only be calculated on `ete3.ClusterTree` objects", file=sys.stderr)

    # Eigengenes
    def eigengenes(self, X=None, mode=0):
        """
        Assumes X.shape follows (obsv_ids, attr_ids)
        mode == 0: return | eigengene pd.DataFrame
        mode == 1: return | eigenvalue pd.Series
        mode == 2: return | (0,1)
        """
        X_cluster = None
        if X is None:
            assert self.data is not None, "If X is None then `data` must be not be None during init"
            X = self.data
        # Updating leaf eaxis
        if self.leaf_axis is None:
            axis_0_overlap = len(set(self.linkage_labels) & set(X.index))
            axis_1_overlap = len(set(self.linkage_labels) & set(X.columns))
            leaf_axis = np.argmax([axis_0_overlap, axis_1_overlap])
            self.leaf_axis = leaf_axis
        assert self.leaf_axis in [0,1], "Could not infer leaf axis.  Please specify the leaf_axis when instantiating the model"

        d_cluster_eigengenes = dict()
        d_cluster_eigenvalues = dict()
        for cluster, leaves in self.get_clusters(grouped=True).iteritems():
            if self.leaf_axis == 0:
                X_cluster = X.loc[leaves,:].T
            if self.leaf_axis == 1:
                X_cluster = X.loc[:,leaves]
            self.pca_model = PrincipalComponentAnalysis(X_cluster, n_components=-1)
            d_cluster_eigengenes[cluster] = pd.Series(self.pca_model.projection_["PC.1"], index=X_cluster.index, name=cluster)
            d_cluster_eigenvalues[cluster] = self.pca_model.explained_variance_ratio_["PC.1"]

        if mode == 0:
            df_eigengenes = pd.DataFrame(d_cluster_eigengenes)
            df_eigengenes.columns.name = "cluster"
            return df_eigengenes
        if mode == 1:
            return pd.Series(d_cluster_eigenvalues)
        if mode == 2:
            return pd.DataFrame(d_cluster_eigengenes), pd.Series(d_cluster_eigenvalues)


    # Intramodular connectivity
    def intramodular_connectivity(self, kernel="auto"):
        if kernel == "auto":
            kernel = 1 - self.df_dism
        if is_query_class(kernel, "Symmetric"):
            kernel = kernel.to_dense()
        return intramodular_connectivity(kernel, clusters=self.get_clusters()[kernel.index])

    # ============
    # Conversion
    # ============
    def is_cluster_node(self, node):
        return node.name in self.cluster_attributes["node"].values

    def as_tree(self, format="ete3", collapse_clusters=False, prefix_cluster="cluster_"):
        accepted_formats = {"ete3", "skbio"}
        assert format in accepted_formats, f"Available formats must be one of the following: {accepted_formats}"
        if format == "ete3":
            if not collapse_clusters:
                return self.tree
            else:
                print("Warning: Collapsing clusters is experimental and should be checked manually before propogating downstream", file=sys.stderr)
                cluster_node = self.cluster_attributes["node"]
                node_cluster = pd.Series(cluster_node.index, index=cluster_node.values)

                if prefix_cluster is None:
                    prefix_cluster = ""

                newick = self.tree.write(is_leaf_fn=self.is_cluster_node, format=0)
                # tree_clusters = self.tree_type(newick)
                # name_ete_nodes(tree_clusters, node_prefix="y")
                tree_clusters = create_tree(newick=newick, into=self.tree_type, force_bifuraction=True)

                for node in tree_clusters.traverse():
                    if self.is_cluster_node(node):
                        id_cluster = node_cluster[node.name]
                        node.name = "{}{}".format(prefix_cluster, id_cluster)
                        for field in  self.cluster_attributes.columns:
                            node.add_feature(field, self.cluster_attributes.loc[id_cluster,field])

                return tree_clusters

        if format == "skbio":
            assert not collapse_clusters, "Cannot currently `collapse_clusters` with `skbio` trees.  Will be available in future versions."
            tree_skbio = skbio.TreeNode.from_linkage_matrix(self.Z, self.linkage_labels)
            tree_skbio.name = self.name
            return tree_skbio


    def as_graph(self, graph=None, weight_func=None):
        # Get graph
        if graph is None:
            graph = nx.OrderedDiGraph(name=self.name)

        # Construct graph
        if weight_func is not None:
            for node in self.tree.traverse():
                for child in node.get_children()[::-1]:
                    branch_length = self.tree.get_distance(node, child)
                    graph.add_edge(node.name, child.name, weight=weight_func(branch_length))
        else:
            for node in self.tree.traverse():
                for child in node.get_children()[::-1]:
                    branch_length = self.tree.get_distance(node, child)
                    graph.add_edge(node.name, child.name, weight=branch_length)
        # Add clusters
        if hasattr(self, "clusters"):
            for (leaf, cluster_data) in self.leaf_attributes.iterrows():
                graph.node[leaf].update(cluster_data.to_dict())
        return graph


# Hierarchical Topology
class Topology(object):
    def __init__(self, X=None, y=None, name=None, class_type=None, attr_type=None, obsv_type=None, metric="euclidean", linkage_method="ward", class_colors="gray", metadata=dict(), initialize=True, axis=0, verbose=True, label_explained_variance_ratio="EV Ratio"):
        self.X = X
        self.y = y
        self.name = name
        self.metric = metric
        self.linkage_method = linkage_method
        self.axis = axis
        self.class_type = class_type
        self.attr_type = attr_type
        self.obsv_type = obsv_type
        self.verbose = verbose
        self.tracks = OrderedDict()
        self.class_colors = class_colors
        self.classes = sorted(self.y.unique())
        self.label_explained_variance_ratio = label_explained_variance_ratio
        self.__synthesized__ = datetime.datetime.utcnow()
        self.metadata = {
            "name":name,
            "metric":metric,
            "linkage_method":linkage_method,
            "dimensions":X.shape if X is not None else None,
            "class_type":class_type,
            "classes":self.classes,
            "attr_type":attr_type,
            "obsv_type":obsv_type,
            "axis":axis,
            "verbose":verbose,
            "synthesized":self.__synthesized__.strftime("%Y-%m-%d %H:%M:%S"),
        }
        self.metadata.update(metadata)
        if initialize:
            condition_1 = X is not None
            condition_2 = y is not None
            if all([condition_1, condition_2]):
                self.compute(overwrite=True)
            else:
                print("X and y must be given to run computation. Returning uninitialized object", file=sys.stderr)

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}({str(self.metadata)[1:-1]})"
    def copy(self):
        return copy.deepcopy(self)

    def _check_proceed(self, attribute, overwrite):
        proceed = True
        if hasattr(self, attribute):
            condition_1 = getattr(self, attribute) is not None
            condition_2 = not overwrite
            if all([condition_1, condition_2]):
                proceed = False
        return proceed

    def eigendecomposition(self, overwrite=False):
        assert self.X is not None, "`X` can not be None to do `eigendecomposition`"
        proceed = self._check_proceed(attribute="eigenprofiles", overwrite=overwrite)
        if proceed:
            self.eigenprofiles, self.eigenvalues = eigenprofiles_from_data(self.X, y=self.y, class_type=self.class_type, mode=None)
        return self

    def dissimilarity(self, overwrite=False):
        assert hasattr(self, "eigenprofiles"), "Must compute `eigendecomposition` before calculating `dissimilarity`"
        proceed = self._check_proceed(attribute="dism", overwrite=overwrite)
        if proceed:
            self.dism = pairwise(self.eigenprofiles, metric=self.metric, axis=0)
            self.labels = self.dism.index
        return self

    def cluster(self, overwrite=False):
        assert hasattr(self, "dism"), "Must compute `dissimilarity` before calculating `cluster`"
        proceed = self._check_proceed(attribute="Z", overwrite=overwrite)
        if proceed:
            self.Z = dism_to_linkage(self.dism, method=self.linkage_method)
        return self

    def dendrogram(self, overwrite=False):
        assert hasattr(self, "Z"), "Must compute `cluster` before calculating `dendrogram`"
        proceed = self._check_proceed(attribute="leaves", overwrite=overwrite)
        if proceed:
            dendrogram_data = sp_hierarchy.dendrogram(self.Z, no_plot=True, labels=self.labels)
            self.leaves = dendrogram_data["ivl"]
            self.icoord = np.asarray(dendrogram_data["icoord"])
            self.dcoord = np.asarray(dendrogram_data["dcoord"])
            self.num_leaves = len(dendrogram_data["ivl"])
            self.leaf_positions = pd.Series(np.arange(5, self.num_leaves * 10 + 5, 10), index=self.leaves)
            self.add_track(name=self.label_explained_variance_ratio, data=self.eigenvalues, color=self.class_colors, plot_kws={"linewidth":1.382})
            self.add_track(name="Size", data=self.y.value_counts(), color=self.class_colors, plot_kws={"linewidth":1.382})
        return self
    def convert_to_tree(self, node_prefix="y", overwrite=False):
        assert hasattr(self, "leaves"), "Must compute `dendrogram` before converting to `ete3.Tree` object`"
        proceed = self._check_proceed(attribute="tree", overwrite=overwrite)
        if proceed:
            newick = linkage_to_newick(self.Z, self.labels)
            root_node = f"{node_prefix}1"
            self.tree = ete3.Tree(newick=newick, name=root_node)
            node_index = int(root_node[-1]) + 1
            self.submodels = [root_node]
            for node in self.tree.iter_descendants():
                if not node.is_leaf():
                    node.name = f"{node_prefix}{node_index}"
                    self.submodels.append(node.name)
                    node_index += 1
            self.internal_node_positions = self._get_internal_nodes_positions(self.icoord, self.dcoord, self.leaf_positions, self.tree)
        return self
    def convert_to_graph(self, overwrite=False):
        # Could use `ete_to_nx` here instead
        assert hasattr(self, "tree"), "Must compute `tree` before converting to `nx.OrderedDiGraph` object`"
        proceed = self._check_proceed(attribute="graph", overwrite=overwrite)
        if proceed:
            self.graph = nx.OrderedDiGraph(name=self.name)
            for node in self.tree.traverse():
                for child in node.get_children()[::-1]:
                    self.graph.add_edge(node.name, child.name)
        return self


    def compute(self, overwrite=True):
        start_time = time.time()
        if self.verbose:
            print("===============================", file=sys.stdout)
            print("Running computational pipeline:", file=sys.stdout)
            print("===============================", file=sys.stdout)
        for computation_step in ["eigendecomposition", "dissimilarity", "cluster", "dendrogram", "convert_to_tree", "convert_to_graph"]:
            if self.verbose: print(computation_step, format_duration(start_time), sep="\t", file=sys.stdout)
            getattr(self,computation_step)(overwrite=overwrite)
        return self
    # ==========
    # Conversion
    # ==========

    # Relabel the submodel nodes
    def relabel_submodels(self, mapping):
        assert isinstance(mapping, Mapping), "`mapping` must be a type of dictionary"
        assert hasattr(self, "tree"), "Must create tree first"


        if set(mapping.values()) >= set(self.submodels):
            if self.verbose:
                print("Skipping relabeling because this dictionary has already been used to relabel the original node/submodel names", file=sys.stderr)
        else:
            assert set(mapping) >= set(self.submodels), f"`mapping` must have a key:value pair for each submodel node: {self.submodels}"
            for i, submodel_node in enumerate(self.submodels):
                node = self.tree.search_nodes(name=submodel_node)[0]
                node.name = self.submodels[i] = mapping[submodel_node]
            if hasattr(self, "graph"):
                self.graph = nx.relabel_nodes(self.graph, mapping, copy=True)
        return self

    # =======
    # Get data
    # =======
    # Get text representation of tree
    def get_ascii(self, show_internal=True, compact=False):
        """
        Get the ascii text tree
        """
        attributes=None
        assert hasattr(self, "tree"), "Must create tree first"
        return self.tree.get_ascii(show_internal=show_internal, compact=compact, attributes=attributes)

    # Get target paths
    def get_paths(self, target=None):
        assert hasattr(self,"graph"), "Must create graph first"

        root = self.tree.name
        if target is None:
            paths = list()
            for target in self.classes:
                paths.append(nx.shortest_path(self.graph, source=root, target=target))
            return paths
        else:
            assert target in self.classes, f"{target} is not in `self.classes`"
            return nx.shortest_path(self.graph, source=root, target=target)

    # Get target matrix
    def get_target_matrix(self, directory=None):
        if directory is not None:
            if type(directory) == str:
                directory = pathlib.Path(directory)
            os.makedirs(directory, exist_ok=True)
        # Nested
        def _get_next_node(current_node, target, descendant_classes):
            path = nx.shortest_path(self.graph, source=current_node, target=target)
            next_node = path[1]
            return next_node
        # Y target matrix
        Y = dict()
        # Iterate through submodel nodes
        for submodel in self.submodels:
            # Get the node object
            node = self.tree.search_nodes(name=submodel)[0]
            # Get the child nodes
            # Get the classes that are lower in the hierarchy
            descendant_classes = node.get_leaf_names()
            # Get the training observations that are in the hierarchy
            y_mask = self.y.map(lambda x: x in descendant_classes)
            y_subset = self.y[y_mask]
            # Path
            y_next = y_subset.map(lambda target:_get_next_node(node.name, target, descendant_classes))
            Y[submodel] = y_next
            # Write training files to a directory
            if directory is not None:
                directory_submodel = os.path.join(directory,submodel)
                os.makedirs(directory_submodel, exist_ok=True)
                self.X.loc[y_next.index,:].to_pickle(os.path.join(directory_submodel, "X.pbz2"), compression="bz2")
                self.y.to_frame(self.class_type).loc[y_next.index,:].to_pickle(os.path.join(directory_submodel, "y.pbz2"), compression="bz2")
        Y = pd.DataFrame(Y).loc[self.X.index,:]
        return Y

    def _get_internal_nodes_positions(self, icoord,dcoord,leaf_positions, tree):
        # Get x positions for each leaf
        leaf_positions = leaf_positions.to_dict()
        x_position_leaf = {v:k for k,v in leaf_positions.items()}
        # Sort in ascending order of height so we are working our way up from the leaves to the root
        idx_ascending_tree = np.argsort(dcoord[:,1])

        # Dynamically set the internal node names as we work our way up so the internal nodes are treated as leaves
        internal_node_positions = dict()
        for i_data, d_data in zip(icoord[idx_ascending_tree], dcoord[idx_ascending_tree]):
            leaf_positions_in_branch = set(i_data) & set(x_position_leaf.keys())
            leaves = [*map(lambda x: x_position_leaf[x], leaf_positions_in_branch)]
            node_name = tree.get_common_ancestor(leaves).name
            position = i_data
            x_position_leaf[np.mean(i_data)] = node_name
            internal_node_positions[node_name] = (np.mean(i_data), np.max(d_data))
        return internal_node_positions

    # ======
    # Tracks
    # ======
    def add_track(self, name, data, alpha=0.618, color="black", pad=0.25, percent_of_figure = 20, ylim=None, plot_type="bar", spines=["top", "bottom", "left", "right"], fillna=0, plot_kws=dict(), label_kws=dict(),  bar_width="auto", check_missing_values=False, convert_to_hex=False, horizontal_kws=dict(), horizontal_lines=[]):
        """
        type(track) should be a pd.Series
        """


        # Keywords
        if plot_type == "bar":
            _plot_kws = {"linewidth":1.382, "alpha":alpha, "width":max(5,self.num_leaves/30) if bar_width == "auto" else bar_width}
        elif plot_type == "area":
            _plot_kws = {"linewidth":1.382, "alpha":alpha}
        else:
            _plot_kws = {}
        _plot_kws.update(plot_kws)

        _horizontal_kws = {"linewidth":1, "linestyle":"-", "color":"black"}
        _horizontal_kws.update(horizontal_kws)

        _label_kws = {"fontsize":15, "rotation":0, "horizontalalignment":"right"}
        _label_kws.update(label_kws)

        # Color of the track
        if color is None:
            color = "black"
        if isinstance(color,str):
            color = pd.Series([color]*len(self.leaves), index=self.leaves)
        if is_dict(color):
            color = pd.Series(color)
        if convert_to_hex:
            _sampleitem_ = color[0]
            assert mpl.colors.is_color_like(_sampleitem_), f"{_sampleitem_} is not color-like"
            if is_rgb_like(_sampleitem_):
                color = color.map(rgb2hex)
        # Check for missing values
        if check_missing_values:
            assert set(data.index) == set(self.leaves), "`data.index` is not the same set as `self.leaves`"
            assert set(color.index) == set(self.leaves), "`color.index` is not the same set as `self.leaves`"
        if plot_type in ["area","line"]:
            if color.nunique() > 1:
                print("Warning: `area` and `line` plots can only be one color.  Defaults to first color in iterable.", file=sys.stderr)
        # Horizontal lines
        if is_number(horizontal_lines):
            horizontal_lines = [horizontal_lines]

        # Convert to ticks
        data = data[self.leaves]
        data.index = data.index.map(lambda x:self.leaf_positions[x])
        color = color[self.leaves]
        color.index = color.index.map(lambda x:self.leaf_positions[x])
        # Store data
        self.tracks[name] = {"data":data, "color":color, "pad":pad, "size":f"{percent_of_figure}%", "plot_type":plot_type, "plot_kws":_plot_kws, "label_kws":_label_kws, "spines":spines, "ylim":ylim, "horizontal_kws":_horizontal_kws, "horizontal_lines":horizontal_lines}

        return self
    # ============
    # Plotting
    # ============
    def _plot_dendrogram_mpl(self, icoord, dcoord, ivl, xlim, ylim, color, ax, distance_label, fig_kws, line_kws, distance_tick_kws):
        # Plot dendrogram
        for xs, ys in zip(icoord, dcoord):
            ax.plot(xs, ys,  color, **line_kws)
        # Adjust ticks
        ax.set_xticklabels([])
        ax.set_ylim(ylim)
        ax.set_xlim(xlim)
        #Distance label
        if distance_label is not None:
            ax.set_ylabel(distance_label, **distance_tick_kws)
        # Append axes objects
        self.axes.append( ax )

    def _plot_nodes_mpl(self, icoord, dcoord, ax, pad,annot_kws=dict()):



        ylim = ax.get_ylim()

        for name, (x,y) in self.internal_node_positions.items():
            plt.plot(x, y, "")
            plt.annotate(name, (x, y),
                         **annot_kws)
        ax.set_ylim(min(ylim), max(ylim)+0.0382*pad)
        return ax


    def _plot_track_mpl(self, name, divider, xlim, ylim, show_track_ticks, show_track_labels):
        ax = divider.append_axes("bottom", size=self.tracks[name]["size"], pad=self.tracks[name]["pad"])
        data = self.tracks[name]["data"]
        c = self.tracks[name]["color"]
        kws = self.tracks[name]["plot_kws"]


        if self.tracks[name]["plot_type"] == "bar":
            ax.bar(data.index, data.values, color=c, **kws)
            kws_outline = kws.copy()
            del kws_outline["alpha"]
            ax.bar(data.index, data.values, color="none", edgecolor=c, **kws_outline)
        else:
            if self.tracks[name]["plot_type"] == "area":
                data.plot(kind="area", color=c.values, ax=ax, **kws)
                data.plot(kind="line", color=c.values, alpha=1,  ax=ax, **kws)
            else:
                data.plot(kind=self.tracks[name]["plot_type"], color=c.values, ax=ax,  **kws)

        if len(self.tracks[name]["horizontal_lines"]):
            for y_pos in self.tracks[name]["horizontal_lines"]:
                ax.axhline(y_pos, **self.tracks[name]["horizontal_kws"])
        # Spines
        for spine in ["top", "bottom", "left", "right"]:
            if spine not in self.tracks[name]["spines"]:
                ax.spines[spine].set_visible(False)
        ax.set_xlim(xlim)
        ax.set_xticklabels([])
        if self.tracks[name]["ylim"] is not None:
            ax.set_ylim(ylim)
        if not show_track_ticks:
            ax.set_yticklabels([])
        if show_track_labels:
            ax.set_ylabel(name, **self.tracks[name]["label_kws"])
        return ax


    # Master plot method
    def plot_topology(self,
             branch_color="auto",
             leaf_font_size=15,
             leaf_rotation=0,
             style="seaborn-white",
             figsize=(13,5),
             title=None,
             ax=None,
             legend=False,
             fig_kws=dict(),
             line_kws=dict(),
             leaf_tick_kws=dict(),
             distance_tick_kws=dict(),
             distance_label="Distance",
             title_kws=dict(),
             title_position=0.95,
             show_class_size=False,
             show_explained_variance_ratio=True,
             annot_kws=dict(),
             # bar_kws=dict(),
            box_facecolor="whitesmoke",
              box_edgecolor="black",
                      box_style="round",
                      box_linewidth=1,
                      box_pad=0.25,
                      box_alpha=0.85,
            ):

        # Initialize
        self.axes = list()
        orientation = "top" # Does not work with bottom, left, or right yet 2018-June-26

        # Limits
                # Independent variable plot width
        tree_width = len(self.leaves) * 10
        xlim = (0,tree_width)
            # Depenendent variable plot height
        maximum_height = max(self.Z[:, 2])
        tree_height = maximum_height + maximum_height * 0.05
        ylim = (0,tree_height)

        # Keywords
        _fig_kws = {"figsize":figsize}
        _fig_kws.update(fig_kws)
        _line_kws = {"linestyle":"-", "linewidth":1.382}
        _line_kws.update(line_kws)
        # _bar_kws = {"linewidth":1.382}
        # _bar_kws.update(bar_kws)
        _distance_tick_kws = {"fontsize":15, "rotation":90}
        if _distance_tick_kws["rotation"] == 0:
            _distance_tick_kws["horizontalalignment"] = "right"
        _distance_tick_kws.update(distance_tick_kws)
        _leaf_tick_kws = {"fontsize":15, "rotation":leaf_rotation}
        _leaf_tick_kws.update(leaf_tick_kws)
        _title_kws = {"fontsize":18, "fontweight":"bold","y":title_position}
        _title_kws.update(title_kws)
        _annot_kws = dict(fontsize=15, fontweight="bold", xytext=(10,15), textcoords='offset points', verticalalignment='top', horizontalalignment='center', bbox= dict(boxstyle=box_style, fc=box_facecolor, ec=box_edgecolor, linewidth=box_linewidth,  alpha=box_alpha))
        _annot_kws.update(annot_kws)
        # Branch colors
        if branch_color == "auto":
            branch_color = {True:"white", False:"black"}[style == "dark_background"]


        # Plotting
        with plt.style.context(style):
            if ax is None:
                fig, ax = plt.subplots(**_fig_kws)
            else:
                fig = plt.gcf()

            # Arguments for dendrogram plotting
            args_dendro = dict(
                icoord=self.icoord,
                dcoord=self.dcoord,
                ivl=self.leaves,
                color=branch_color,
                xlim=xlim,
                ylim=ylim,
                ax=ax,
                fig_kws=_fig_kws,
                line_kws=_line_kws,
                distance_tick_kws=_distance_tick_kws,
                distance_label=distance_label,
            )

            self._plot_dendrogram_mpl(**args_dendro)
            self._plot_nodes_mpl(icoord=self.icoord, dcoord=self.dcoord, ax=ax, pad=box_pad,annot_kws=_annot_kws)

            # Add tracks
            divider = make_axes_locatable(ax)
            additional_tracks = list()
            if show_explained_variance_ratio:
                additional_tracks.append(self.label_explained_variance_ratio)
            if show_class_size:
                additional_tracks.append("Size")

            if len(additional_tracks) > 0:
                for name in additional_tracks:
                    if "edgecolor" in self.tracks[name]["plot_kws"]:
                        if self.tracks[name]["plot_kws"]["edgecolor"] == "auto":
                            self.tracks[name]["plot_kws"]["edgecolor"] = {True:"white", False:"black"}[style == "dark_background"]
                    args_track = dict(
                        name=name,
                        divider=divider,
                        xlim=xlim,
                        ylim=ylim,
                        show_track_ticks=False,
                        show_track_labels=True,
                    )
                    ax_track = self._plot_track_mpl( **args_track)
                    self.axes.append(ax_track)

            # Leaves
            [*map(lambda ax_query: ax_query.get_xaxis().set_visible(False), self.axes[:-1])]
            self.axes[-1].set_xticks(self.leaf_positions.values)
            self.axes[-1].set_xticklabels(self.leaves, **_leaf_tick_kws)
            self.axes[-1].xaxis.set_ticks_position("bottom")

            # Remove grids
            [*map(lambda ax_query: ax_query.grid(False), self.axes)]

            # Title
            if title is None:
                title = self.name
            if title is not None:
                fig.suptitle(title, **_title_kws)
            return fig, ax



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys, time, multiprocessing

import pandas as pd
import numpy as np

# Soothsayer
from ..r_wrappers import *
from soothsayer.utils import *

# ==============================================================================
# R Imports
# ==============================================================================
from rpy2 import robjects, rinterface
from rpy2.robjects.packages import importr
from rpy2.rinterface import RRuntimeError
from rpy2.robjects import pandas2ri
pandas2ri.activate()
R = robjects.r
NULL = robjects.rinterface.NULL
rinterface.set_writeconsole_regular(None)

# R packages
wgcna = R_package_retrieve("WGCNA")

def TOMsimilarity(adjacency,  TOMType="unsigned",  TOMDenom="min"):
    """
    WGCNA: TOMsimilarity
        function (adjMat, TOMType = "unsigned", TOMDenom = "min", verbose = 1,
                   indent = 0)
    """
    assert np.all(adjacency.index == adjacency.columns), "pd.DataFrame must have symmetric labels for index and columns"
    nodes = adjacency.index
    rDF_adj = pandas2ri.py2ri(adjacency)
    rDF_tom = wgcna.TOMsimilarity(R["as.matrix"](rDF_adj), TOMType=TOMType, TOMDenom=TOMDenom, verbose=0)
    Ar_tom = pandas2ri.ri2py(rDF_tom)
    return pd.DataFrame(Ar_tom, index=nodes, columns=nodes)

def bicor(X, axis=1, n_jobs=-1):
    """
    WGCNA: bicor
        function (x, y = NULL, robustX = TRUE, robustY = TRUE, use = "all.obs",
                   maxPOutliers = 1, qu <...> dian absolute deviation, or zero variance."))
    """
    if n_jobs == -1:
        n_jobs = multiprocessing.cpu_count()

    if axis == 1: X_copy = X.copy()
    if axis == 0: X_copy = X.copy().T
    labels = X_copy.columns
    rDF_sim = wgcna.bicor(pandas2ri.py2ri(X_copy))
    df_bicor = pd.DataFrame(pandas2ri.ri2py(rDF_sim), index=labels, columns=labels)
    return df_bicor

def pickSoftThreshold_fromSimilarity(df_adj, query_powers):
    # Run pickSoftThreshold.fromSimilarity
    rDF_scalefreetopology = wgcna.pickSoftThreshold_fromSimilarity(R["as.matrix"](pandas2ri.py2ri(df_adj)), powerVector = query_powers, verbose=0, moreNetworkConcepts=True)[1]
    return pandas2ri.ri2py(rDF_scalefreetopology)



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time
from collections import OrderedDict, defaultdict
# from scipy.cluster import hierarchy as sp_hierarchy
import networkx as nx
import pandas as pd
import ete3
import skbio
from ..transmute.conversion import ete_to_skbio, nx_to_ete, name_ete_nodes

__all__ = ["create_tree"]
__all__ = sorted(__all__)

# ==============================================================================
# Trees
# ==============================================================================
# Create ete3 tree or skbio
def create_tree(
        # Base
        newick=None,
        name=None,
        format=0,
        dist=1.0,
        support=1.0,
        quoted_node_names=False,
        # ClusterTree
        text_array=None,
        fdist=None,
        # PhyloTree
        alignment=None,
        alg_format='fasta',
        sp_naming_function=None,
        # PhyloxmlTree
        phyloxml_clade=None,
        phyloxml_phylogeny=None,
        # Constructor
        node_prefix="y",
        into=ete3.Tree,
        prune=None,
        force_bifuraction=True,
        # Keywords
        tree_kws=dict(),
        bifurcation_kws=dict(recursive=True),
    ):
    """
    Next: Convert to NetworkX
    """
    # Should the tree be converted to skbio
    convert_to_skbio = False
    if into in [skbio.TreeNode]:
        into = ete3.Tree
        convert_to_skbio = True

    # ete3 construction
    if into == ete3.Tree:
        tree = ete3.Tree(newick=newick, format=format, quoted_node_names=quoted_node_names, **tree_kws)
    if into == ete3.ClusterTree:
        if isinstance(text_array, pd.DataFrame):
            text_array = dataframe_to_matrixstring(text_array)
        tree = ete3.ClusterTree(newick=newick, text_array=text_array, fdist=fdist, **tree_kws)
    if into == ete3.PhyloTree:
        tree = ete3.PhyloTree(newick=newick, alignment=alignment, alg_format=alg_format, sp_naming_function=sp_naming_function, format=format, **tree_kws)
    if into == ete3.PhyloxmlTree:
        tree = ete3.PhyloxmlTree(phyloxml_clade=phyloxml_clade, phyloxml_phylogeny=phyloxml_phylogeny,**tree_kws)

    # Set base attributes
    for k,v in dict(name=name, dist=dist, support=support).items():
        setattr(tree, k, v)

    # Prune
    if prune is not None:
        tree.prune(prune)

    # Bifurcation
    if force_bifuraction:
        n_internal_nodes = len([*filter(lambda node:node.is_leaf() == False, tree.traverse())])
        n_leaves = len([*filter(lambda node:node.is_leaf(), tree.traverse())])
        if n_internal_nodes < (n_leaves - 1):
            tree.resolve_polytomy(**bifurcation_kws)

    # Node prefix
    if node_prefix is not None:
        tree = name_ete_nodes(tree, node_prefix=node_prefix)
    if not convert_to_skbio:
        return tree
    # skbio
    else:
        return ete_to_skbio(tree, node_prefix=None)



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, datetime, pathlib
from collections import OrderedDict, defaultdict
from io import StringIO, BytesIO, TextIOWrapper
from ast import literal_eval

# Compression & Serialization
import pickle, gzip, bz2, zipfile

# PyData
import pandas as pd

# Soothsayer
from soothsayer.utils import format_path, infer_compression, is_file_like

# Create file object
def get_file_object(path:str, mode="infer", compression="infer", safe_mode="infer", verbose=True):
    """
    # Should this be in io?
    with get_file_object("./test.txt.zip", mode="infer", verbose=False) as f_read:
    with get_file_object("./test_write.txt.bz2", mode="write", verbose=False) as f_write:
        for line in f_read.readlines():
            line = str(line.strip())
            print(line, file=f_write)
    """
    # Init
    f = None
    file_will_be_written = False

    # Paths
    path = format_path(path)
    path_exists = os.path.exists(path)

    if verbose:
        header = " ".join(["Processing file:", path])
        print("="*len(header), header, "="*len(header), sep="\n", file=sys.stderr)
    if compression == "infer":
        compression = infer_compression(path)
        if verbose:
            print("Inferring compression:", compression, file=sys.stderr)

    # Inferring mode
    if mode == "infer": # Create new function for this? infer_filemode?
        if path_exists:
            mode = "read"
        else:
            mode = "write"
    assert mode != "infer", "The mode should be inferred by this point.  Please specify mode manually."
    assert compression != "infer", "The compression should be inferred by this point.  Please specify compression manually."

    # Generic read write
    if mode in ["write", "read"]:
        if mode == "write":
            mode = "w"
        if mode == "read":
            mode = "r"
        if compression in ["gzip", "bz2"]:
            mode = mode + "b"
        if verbose:
            print("Inferring mode:", mode, file=sys.stderr)

    # Will a file be written?
    if "w" in mode:
        file_will_be_written = True

    # Ensure zip is not being written
    if file_will_be_written:
        assert compression != "zip", "Currently cannot handle writing zip files.  Please use gzip, bz2, or None."
        # Future do this:
        # https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile.open

    # Safe mode
    if safe_mode == "infer":
        if file_will_be_written:
            safe_mode = True
        else:
            safe_mode = False
    assert safe_mode in {True,False}, "Please choose either True or False for `safe_mode`"
    if safe_mode:
        if all([file_will_be_written, path_exists]):
            raise Exception(f"Safe Mode: Please explicitly provide a writeable mode ['w', 'wb', or 'write'] because `{path}` already exists and will be rewritten.")

    # GZIP compression
    if compression == "gzip":
        f = gzip.open(path, mode)
    # BZ2 compression
    if compression == "bz2":
        f = bz2.open(path, mode)
    if compression == "zip":
        filename, ext = os.path.splitext(os.path.split(path)[-1])
        f = zipfile.ZipFile(path,mode).open(filename)
    # No compression
    if f is None:
        return open(path, mode)

    # Reading and writing compressed files
    else:
        return TextIOWrapper(f, encoding="utf-8")


# Text reading wrapper
def read_textfile(path, enum=True, generator=True, mode="read", compression="infer", into=pd.Series):
    """
    2018-May-29
    Edits: 2018-December-27: Added `get_file_object` dependency
    """
    assert mode not in ["w", "wb", "a"], "`mode` should not be in {w, wb, a} because it will overwrite"
    assert compression  in ["infer", "gzip", "bz2", "zip", None], "Valid `compression` types are 'infer', 'gzip', 'bz2', 'zip'"
    # Format file path
    path = format_path(path, str)
    # Get file object
    handle = get_file_object(path=path, mode=mode, compression=compression, safe_mode=False, verbose=False)
    # Subfunctions
    def run_return_list(handle):
        data = handle.read().split("\n")
        handle.close()
        if into == pd.core.series.Series:
            return pd.Series(data, name=path)
        else:
            if enum:
                return into([*enumerate(data)])
            if not enum:
                return into(data)

    def run_return_iterable(handle):
        if enum:
            for i,line in enumerate(handle.readlines()):
                yield i, line.strip()
        if not enum:
            for line in handle.readlines():
                yield line.strip()
        handle.close()

    # Controller
    if generator:
        return run_return_iterable(handle=handle)
    if not generator:
        return run_return_list(handle=handle)
